Package: iTensor
Type: Package
Title: ICA-Based Matrix/Tensor Decomposition
Version: 1.0.5
Authors@R: c(person("Koki", "Tsuyuzaki", role = c("aut", "cre"),
                      email = "k.t.the-answer@hotmail.co.jp"))
Description: Some functions for performing ICA, MICA, Group ICA, and Multilinear ICA are implemented.
    ICA, MICA/Group ICA, and Multilinear ICA extract statistically independent components from single matrix, multiple matrices, and single tensor, respectively.
    For the details of these methods, see the reference section of GitHub README.md <https://github.com/rikenbit/iTensor>.
License: MIT + file LICENSE
Encoding: UTF-8
Depends: R (>= 3.4.0)
URL: https://github.com/rikenbit/iTensor
RoxygenNote: 7.2.3
Suggests: nnTensor, knitr, rmarkdown, testthat (>= 3.0.0)
Config/testthat/edition: 3
Imports: MASS, methods, graphics, utils, stats, rTensor, jointDiag,
        mgcv, einsum, geigen, mixOmics
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-05-11 08:21:22 UTC; koki
Author: Koki Tsuyuzaki [aut, cre]
Maintainer: Koki Tsuyuzaki <k.t.the-answer@hotmail.co.jp>
Repository: CRAN
Date/Publication: 2026-05-11 16:50:02 UTC
Built: R 4.6.0; ; 2026-05-11 19:33:47 UTC; unix

Package: iTensor
Type: Package
Title: ICA-Based Matrix/Tensor Decomposition
Version: 1.0.5
Authors@R: c(person("Koki", "Tsuyuzaki", role = c("aut", "cre"),
                      email = "k.t.the-answer@hotmail.co.jp"))
Description: Some functions for performing ICA, MICA, Group ICA, and Multilinear ICA are implemented.
    ICA, MICA/Group ICA, and Multilinear ICA extract statistically independent components from single matrix, multiple matrices, and single tensor, respectively.
    For the details of these methods, see the reference section of GitHub README.md <https://github.com/rikenbit/iTensor>.
License: MIT + file LICENSE
Encoding: UTF-8
Depends: R (>= 3.4.0)
URL: https://github.com/rikenbit/iTensor
RoxygenNote: 7.2.3
Suggests: nnTensor, knitr, rmarkdown, testthat (>= 3.0.0)
Config/testthat/edition: 3
Imports: MASS, methods, graphics, utils, stats, rTensor, jointDiag,
        mgcv, einsum, geigen, mixOmics
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-05-11 08:21:22 UTC; koki
Author: Koki Tsuyuzaki [aut, cre]
Maintainer: Koki Tsuyuzaki <k.t.the-answer@hotmail.co.jp>
Repository: CRAN
Date/Publication: 2026-05-11 16:50:02 UTC
Built: R 4.5.2; ; 2026-05-11 20:16:33 UTC; unix

# Generated by roxygen2: do not edit by hand

export(CorrIndex)
export(GroupICA)
export(ICA)
export(ICA2)
export(MICA)
export(MultilinearICA)
export(toyModel)
import(mixOmics)
importFrom(MASS,ginv)
importFrom(einsum,einsum)
importFrom(geigen,geigen)
importFrom(graphics,hist)
importFrom(jointDiag,uwedge)
importFrom(methods,is)
importFrom(mgcv,gam)
importFrom(rTensor,as.tensor)
importFrom(rTensor,cs_unfold)
importFrom(rTensor,ttl)
importFrom(rTensor,ttm)
importFrom(stats,cor)
importFrom(stats,cov)
importFrom(stats,dnorm)
importFrom(stats,filter)
importFrom(stats,prcomp)
importFrom(stats,predict)
importFrom(stats,rnorm)
importFrom(stats,rt)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,var)
importFrom(utils,data)
importFrom(utils,str)

## ----data, echo=TRUE----------------------------------------------------------
library("iTensor")
data1 <- iTensor::toyModel("ICA_Type1")
data2 <- iTensor::toyModel("ICA_Type2")
data3 <- iTensor::toyModel("ICA_Type3")
data4 <- iTensor::toyModel("ICA_Type4")
data5 <- iTensor::toyModel("ICA_Type5")
dim(data1$X_observed)
dim(data2$X_observed)
dim(data3$X_observed)
dim(data4$X_observed)
dim(data5$gene) # N < P systems

## ----data2, echo=TRUE, fig.height=4, fig.width=4------------------------------
pairs(data1$X_observed, main="data1")
pairs(data2$X_observed, main="data2")
pairs(data3$X_observed, main="data3")
pairs(data4$X_observed, main="data4")
dim(data5$gene)

## ----setting, echo=TRUE-------------------------------------------------------
allICA <- function(X, J){
    # Classic ICA Methods
    out.FastICA <- ICA(X=X, J=J, algorithm="FastICA")
    out.InfoMax <- ICA(X=X, J=J, algorithm="InfoMax")
    out.ExtInfoMax <- ICA(X=X, J=J, algorithm="ExtInfoMax")
    # Modern ICA Method
    out.JADE <- ICA2(X=X, J=J, algorithm="JADE")
    # Output
    list(out.FastICA=out.FastICA, out.InfoMax=out.InfoMax,
      out.ExtInfoMax=out.ExtInfoMax, out.JADE=out.JADE)
}

CorrIndexAllICA <- function(S, out){
    tmp <- lapply(out, function(x, S){CorrIndex(cor(S, Re(x$S)))}, S=S)
    Name <- gsub("out.", "", names(tmp))
    Value <- round(unlist(tmp), 3)
    names(Value) <- NULL
    cbind(Name, Value)
    tmp <- as.data.frame(cbind(Name, Value))
    colnames(tmp) <- c("Method", "CorrIndex")
    knitr::kable(tmp)
}

## ----ica1, echo=TRUE----------------------------------------------------------
set.seed(123456)
out1 <- allICA(X=data1$X_observed, J=3)

## ----corrindex1, echo=TRUE----------------------------------------------------
CorrIndexAllICA(data1$S_true, out1)

## ----pairs_ica1, echo=TRUE, fig.height=4, fig.width=4-------------------------
pairs(out1$out.FastICA$S, main="FastICA")
pairs(out1$out.InfoMax$S, main="InfoMax")
pairs(out1$out.ExtInfoMax$S, main="ExtInfoMax")
pairs(Re(out1$out.JADE$S), main="JADE")

## ----ica2, echo=TRUE----------------------------------------------------------
set.seed(123456)
out2 <- allICA(X=data2$X_observed, J=3)

## ----corrindex2, echo=TRUE----------------------------------------------------
CorrIndexAllICA(data2$S_true, out2)

## ----pairs_ica2, echo=TRUE, fig.height=4, fig.width=4-------------------------
pairs(out2$out.FastICA$S, main="FastICA")
pairs(out2$out.InfoMax$S, main="InfoMax")
pairs(out2$out.ExtInfoMax$S, main="ExtInfoMax")
pairs(Re(out2$out.JADE$S), main="JADE")

## ----ica3, echo=TRUE----------------------------------------------------------
set.seed(123456)
out3 <- allICA(X=data3$X_observed, J=3)

## ----corrindex3, echo=TRUE----------------------------------------------------
CorrIndexAllICA(data3$S_true, out3)

## ----pairs_ica3, echo=TRUE, fig.height=4, fig.width=4-------------------------
pairs(out3$out.FastICA$S, main="FastICA")
pairs(out3$out.InfoMax$S, main="InfoMax")
pairs(out3$out.ExtInfoMax$S, main="ExtInfoMax")
pairs(Re(out3$out.JADE$S), main="JADE")

## ----ica4, echo=TRUE----------------------------------------------------------
set.seed(123456)
out4 <- allICA(X=data4$X_observed, J=3)

## ----corrindex4, echo=TRUE----------------------------------------------------
CorrIndexAllICA(data4$S_true, out4)

## ----pairs_ica4, echo=TRUE, fig.height=4, fig.width=4-------------------------
pairs(out4$out.FastICA$S, main="FastICA")
pairs(out4$out.InfoMax$S, main="InfoMax")
pairs(out4$out.ExtInfoMax$S, main="ExtInfoMax")
pairs(Re(out4$out.JADE$S), main="JADE")

## ----ica5, echo=TRUE----------------------------------------------------------
library("mixOmics")
set.seed(123456)

# mixOmics's IPCA
res.ipca <- ipca(data5$gene, ncomp=3, mode="deflation")

# iTensor's IPCA
out5 <- ICA2(X=as.matrix(data5$gene), J=3, algorithm="IPCA")

## ----pairs_ica5, echo=TRUE, fig.height=4, fig.width=4-------------------------
pairs(res.ipca$x, main="IPCA (mixOmics)", col=data5$treatment[,"Treatment.Group"], pch=16)
pairs(out5$S, main="IPCA (iTensor)", col=data5$treatment[,"Treatment.Group"], pch=16)

## ----sessionInfo, echo=FALSE--------------------------------------------------
sessionInfo()

#' Multilinear independent component analysis
#'
#'#' The input object is assumed to be a Tensor object defined by rTensor package.
#' In MultilinearICA, ICA function is performed in each mode of the tensor.
#' @param X An rTensor object
#' @param Js A vector to specify the rank in each mode (Default: c(3,3,3))
#' @param modes A vector to specify which modes are decomposed (Default: 1:3)
#' @param algorithm The algorithm to decompose the input tensor in each mode (Default: "FastICA")
#' @return A list containing the result of the decomposition
#' @examples
#' library("rTensor")
#' arrX <- array(runif(10*20*30), dim=c(10,20,30))
#' X <- as.tensor(arrX)
#' Js <- c(2,3,4)
#' out <- MultilinearICA(X, Js=Js)
#' @importFrom rTensor cs_unfold ttl ttm as.tensor
#' @importFrom methods is
#' @export
MultilinearICA <- function(X, Js=c(3,3,3), modes=1:3,
    algorithm=c("FastICA", "InfoMax", "ExtInfoMax")){
    ######################################
    # Argument Check
    ######################################
    .checkMultilinearICA(X, Js, modes)
    algorithm <- match.arg(algorithm)
    ######################################
    # Initialization
    ######################################
    int <- .initMultilinearICA(X, Js, modes)
    As <- int$As
    S <- int$S
    ######################################
    # ICA in each mode
    ######################################
    nrs <- dim(X)
    for(m in seq_along(nrs)){
        if(m %in% modes){
            position <- which(modes == m)
            Xm <- cs_unfold(X, m=m)@data
            As[[m]] <- ICA(X=Xm, J=Js[position], algorithm=algorithm)$A
        }
    }
    # Core tensor
    S <- .projection(X, As, modes=modes)
	# Output
    list(As=As, S=S, Js=Js, algorithm=algorithm)
}

.checkMultilinearICA <- function(X, Js, modes){
    stopifnot(is(X)[1] == "Tensor")
    stopifnot(is.vector(Js))
    stopifnot(length(Js) == length(modes))
    for(m in seq_along(modes)){
        stopifnot(dim(X)[modes[m]] >= Js[m])
    }
    stopifnot(all(.ndim(X) >= length(Js)))
    stopifnot(all(1 <= modes))
    stopifnot(all(.ndim(X) >= length(modes)))
}

.ndim <- function(X){
    length(dim(X))
}

.initMultilinearICA <- function(X, Js, modes){
	# As
	nrs <- dim(X)
    As <- list()
    length(As) <- length(nrs)
    for(m in seq_along(nrs)){
        if(m %in% modes){
            position <- which(modes == m)
            As[[m]] <- matrix(0, nrow=Js[position], ncol=nrs[m])
        }else{
            As[[m]] <- diag(nrs[m])
        }
    }
    # S
	S <- as.tensor(array(0, dim=Js))
	list(As=As, S=S)
}

.projection <- function(X, As, modes=modes){
    out <- X
    for(m in seq_len(.ndim(X))){
        if(m %in% modes){
            out <- ttm(out, ginv(t(As[[m]])), m=m)
        }else{
            out <- ttm(out, t(As[[m]]), m=m)
        }
    }
    out
}

# Simulation Dataset
arrX <- array(runif(10*20*30), dim=c(10,20,30))
X <- as.tensor(arrX)

## Perform MICA against Simulation Datasets
Js <- c(2,3,4)
modes <- 1:3
out <- MultilinearICA(X, Js=Js, algorithm="FastICA")

## Test Hidden Functions
expect_true(is.null(iTensor:::.checkMultilinearICA(X, Js, modes=modes)))
expect_equal(iTensor:::.ndim(X), 3)
expect_true(is.list(iTensor:::.initMultilinearICA(X, Js, modes=modes)))
expect_true("Tensor" %in% is(iTensor:::.projection(X, out$As, modes=modes)))

## Test Output object / type
### Test O-1: Object
expect_identical(is.list(out), TRUE)

### Test O-2: Object Names
expect_identical(names(out),
    c("As", "S", "Js", "algorithm"))

### Test 0-3: As
expect_identical(is.list(out$As), TRUE)
expect_equal(dim(out$As[[1]]), c(Js[1], dim(X)[1]))
expect_equal(dim(out$As[[2]]), c(Js[2], dim(X)[2]))
expect_equal(dim(out$As[[3]]), c(Js[3], dim(X)[3]))

### Test 0-4: S
expect_identical(is(out$S)[1], "Tensor")

expect_equal(dim(out$S)[1], Js[1])
expect_equal(dim(out$S)[2], Js[2])
expect_equal(dim(out$S)[3], Js[3])

### Test 0-5: Js
expect_identical(out$Js, Js)

### Test 0-6: algorithm
expect_identical(out$algorithm, "FastICA")

## Test Error
### Test E-1: X
expect_error(MultilinearICA(arrX, Js=Js))

### Test E-2: Js
expect_error(MultilinearICA(X, Js=c("2", "3", "4")))
expect_error(MultilinearICA(X, Js=c(2, 3, 4, 5)))
expect_error(MultilinearICA(X, Js=c(2, 3)))

### Test E-3: modes
expect_error(MultilinearICA(X, modes=c("1", "2", "3")))
expect_error(MultilinearICA(X, modes=1:4))
expect_error(MultilinearICA(X, modes=c(1, 3)))

### Test E-4: algorithm
expect_error(MultilinearICA(X, algorithm="FastICAA"))

## All Combination of J and modes
expect_error(expect_error(MultilinearICA(X, Js=c(2,3,4), modes=c(1,2,3))))
expect_error(expect_error(MultilinearICA(X, Js=c(2,4), modes=c(1,3))))
expect_error(expect_error(MultilinearICA(X, Js=c(3,4), modes=c(2,3))))
expect_error(expect_error(MultilinearICA(X, Js=2, modes=1)))
expect_error(expect_error(MultilinearICA(X, Js=3, modes=2)))
expect_error(expect_error(MultilinearICA(X, Js=4, modes=3)))

#' Group Independent Component Analysis (GroupICA)
#'
#' The input data is assumed to be a list containing multiple matrices, which share common column.
#' @param Xs A list containing multiple matrices
#' @param J1 Rank parameter to decompose
#' @param J2 Rank parameter used in Calhoun2009
#' @param algorithm Pool algorithm to merge multiple ICA results (Default: pooled)
#' @param ica.algorithm The decomposition algorithm (Default: "FastICA")
#' @param num.iter The number of iterations
#' @param thr The threshold to terminate the iteration (Default: 1E-10)
#' @param verbose Verbose option
#' @return A list containing the result of the decomposition
#' @examples
#' X1 <- matrix(runif(100*200), nrow=100, ncol=200)
#' X2 <- matrix(runif(150*200), nrow=150, ncol=200)
#' Xs <- list(X1=X1, X2=X2)
#' out <- GroupICA(Xs, J1=5)
#' @importFrom jointDiag uwedge
#' @importFrom utils str
#' @importFrom stats prcomp
#' @export
GroupICA <- function(Xs, J1, J2 = J1,
    algorithm = c("pooled", "Calhoun2009", "Pfister2018"),
    ica.algorithm = c("FastICA", "InfoMax", "ExtInfoMax",
        "JADE", "AuxICA1", "AuxICA2", "IPCA", "SIMBEC",
        "AMUSE", "SOBI", "FOBI", "ProDenICA", "RICA"),
    num.iter = 30, thr = 1E-10, verbose = FALSE){
    ######################################
    # Argument Check
    ######################################
    algorithm <- match.arg(algorithm)
    ica.algorithm <- match.arg(ica.algorithm)
    .checkGroupICA(Xs, J1, J2, algorithm, num.iter, thr, verbose)
    l <- length(Xs)
    if(algorithm == "pooled"){
        out <- .pooled(Xs, J1, J2, ica.algorithm, num.iter, thr, verbose)
    }
    if(algorithm == "Calhoun2009"){
        out <- .Calhoun2009(Xs, J1, J2, ica.algorithm, num.iter, thr, verbose)
    }
    if(algorithm == "Pfister2018"){
        out <- .Pfister2018(Xs, J1, J2, ica.algorithm, num.iter, thr, verbose)
    }
    # Output
    list(A = out$A, Ss = out$Ss, RecError = out$RecError, RelChange = out$RelChange)
}

.pooled <- function(Xs, J1, J2, ica.algorithm, num.iter, thr, verbose){
    Xpooled <- do.call(rbind, Xs)
    if(ica.algorithm %in% c("FastICA", "InfoMax", "ExtInfoMax")){
        out <- ICA(Xpooled, J = J1, algorithm = ica.algorithm,
            num.iter = num.iter, thr = thr, verbose = verbose)
    }else{
    out <- ICA2(Xpooled, J = J1, algorithm = ica.algorithm,
        num.iter = num.iter, thr = thr, verbose = verbose)
    }
    A <- out$A
    group_index <- cumsum(as.numeric(unlist(lapply(
        Xs, function(x){
            l <- numeric(nrow(x))
            l[1] <- 1
            l
        }))))
    Ss <- list()
    for(i_group in seq_len(max(group_index))){
        Ss[[i_group]] <- out$S[group_index == i_group, ]
    }
    RecError <- out$RecError
    RelChange <- out$RelChange
    list(A = A, Ss = Ss, RecError = RecError, RelChange = RelChange)
}

.Calhoun2009 <- function(Xs, J1, J2, ica.algorithm, num.iter, thr, verbose){
    # The variable names are the same as those in the paper.
    # K is the number of samples in each group
    Ks <- lapply(Xs, function(x){
        nrow(x)
    })
    # L is the size of time dimension following reduction.
    L <- J2
    # V is the number of input components
    V <- ncol(Xs[[1]])
    # M is the number of groups
    M <- length(Xs)
    # N is the number of components to be estimated
    N <- J1
    if(verbose){
        print(paste("Ks:", Ks))
        print(paste("L:", L))
        print(paste("V:", V))
        print(paste("M:", M))
        print(paste("N:", N))
    }
    # each element of Ys is a K x V matrix
    Ys <- lapply(Xs, function(x){
        # scale adds unnecessary attributes, so we cut them off with `[,]`.
        scale(x, center=TRUE, scale=FALSE)[,]
    })
    if(verbose){
        print(paste("Ys:", lapply(Ys, dim)))
        print("contents of Ys:")
        print(str(Ys))
    }
    Y_prcomps <- lapply(Ys, function(Y){
        prcomp(t(Y), center=TRUE, scale=FALSE)
    })
    # each element of F_inv_Ys is a L x V matrix
    F_inv_Ys <- lapply(Y_prcomps, function(Y_prcomp){
        t(Y_prcomp$x[, seq_len(L)])
    })
    if(verbose){
        print(paste("F_inv_Ys:", lapply(F_inv_Ys, dim)))
        print("contents of F_inv_Ys:")
        print(str(F_inv_Ys))
    }
    # each element of F_invs is a L x K matrix
    F_invs <- lapply(Y_prcomps, function(Y_prcomp){
        t(Y_prcomp$rotation[, seq_len(L)])
    })
    if(verbose){
        print(paste("F_invs:", lapply(F_invs, dim)))
        print("contents of F_invs:")
        print(str(F_invs))
    }
    # F_inv_Y_concatenated ... (LM x V)
    F_inv_Y_concatenated <- do.call(rbind, F_inv_Ys)
    if(verbose){
        print("F_inv_Y_concatenated:")
        print(dim(F_inv_Y_concatenated))
    }
    # G_inv ... (N x LM)
    G_inv <- t(prcomp(t(F_inv_Y_concatenated), center=TRUE, scale=FALSE)$rotation[, seq_len(N)])
    if(verbose){
        print("G_inv:")
        print(dim(G_inv))
        print("contents of G_inv:")
        print(G_inv)
    }
    X <- t(prcomp(t(F_inv_Y_concatenated), center=TRUE, scale=FALSE)$x[, seq_len(N)])
    if(verbose){
        print("X:")
        print(dim(X))
        print("contents of X:")
        print(X)
    }
    if(ica.algorithm %in% c("FastICA", "InfoMax", "ExtInfoMax")){
        X.ica <- ICA(X, J = V, algorithm = ica.algorithm,
            num.iter = num.iter, thr = thr, verbose = verbose)
    }else{
        X.ica <- ICA2(X, J = V, algorithm = ica.algorithm,
            num.iter = num.iter, thr = thr, verbose = verbose)
    }
    # A ... N x N
    A <- X.ica$A
    if(verbose){
        print("A:")
        print(dim(A))
        print("contents of A:")
        print(A)
    }
    # S ... N x V
    S <- X.ica$S
    if(verbose){
        print("S:")
        print(dim(S))
        print("contents of S:")
        print(S)
    }
    # G ... LM x N
    G <- ginv(G_inv)
    if(verbose){
        print("G:")
        print(dim(G))
        print("contents of G:")
        print(G)
    }
    # each element of Gs is a L x N matrix
    Gs <- list()
    for(i_group in seq_len(M)){
        Gs[[i_group]] <- G[seq((i_group - 1) * L + 1, i_group * L), ]
    }
    if(verbose){
        print(paste("Gs:", lapply(Gs, dim)))
        print("contents of Gs:")
        print(str(Gs))
    }
    # each element of Ss is a N x V matrix
    Ss <- list()
    for(i_group in seq_len(M)){
        Ss[[i_group]] <- ginv(Gs[[i_group]] %*% A) %*% F_inv_Ys[[i_group]]
    }
    if(verbose){
        print(paste("Ss:", lapply(Ss, dim)))
        print("contents of Ss:")
        print(str(Ss))
    }
    # calculate ICA time courses
    # each element of FGAs is a K x N matrix
    FGAs <- lapply(seq_len(M), function(i_group){
        # F_invs[[i_group]] ... L x K
        # Gs[[i_group]] ... L x N
        # A ... N x N
        ginv(F_invs[[i_group]]) %*% Gs[[i_group]] %*% A
    })
    if(verbose){
        print(paste("FGAs:", lapply(FGAs, dim)))
        print("contents of FGAs:")
        print(str(FGAs))
    }
    # Because F_i G_i A is considered to be the original signal,
    # we output it as S_i.
    # Note that S_i is N x V.
    Ss <- FGAs
    RecError <- NULL
    RelChange <- NULL
    list(A = A, Ss = Ss, RecError = RecError, RelChange = RelChange)
}

.Pfister2018 <- function(Xs, J1, J2, ica.algorithm, num.iter, thr, verbose){
    ######################################
    # Initialization (e.g. Whiteniing)
    ######################################
    int <- .initGroupICA(Xs)
    X <- int$X
    g <- int$g    # group index
    Pg <- int$Pg  # group-wise partition
    M <- int$M    # empty list
    whitening_result <- .whitening2(X)
    X <- whitening_result$XWhiten
    X <- X[, seq_len(J1)]
    groups <- unique(g)
    for(g_i in groups){
        partitions <- unique(Pg[g == g_i])
        for(Pg_i in partitions){
            e <- which(g == g_i & Pg == Pg_i)
            # which(g != g_i | Pg != Pg_i) is not necessarily correct.
            # which(g != g_i & Pg == Pg_i) may be correct.
            # e_complement <- which(g != g_i | Pg != Pg_i)
            e_complement <- which(g != g_i | Pg == Pg_i)
            M[[length(M) + 1]] <- cov(X[e, ]) - cov(X[e_complement, ])
            M[[length(M) + 1]] <- cov(X[e_complement, ]) - cov(X[e, ])
        }
    }
    n_eigen_matrices_diagonalized <- length(M)
    M_stacked <- array(0.0, dim = c(J1, J1, n_eigen_matrices_diagonalized))
    for(i in seq_len(n_eigen_matrices_diagonalized)){
        M_stacked[, , i] <- M[[i]]
    }
    A <- .ApproximateJointDiagonalizer(M_stacked)
    # Output
    S <- X %*% A
    group_index <- cumsum(as.numeric(unlist(lapply(
        Xs, function(x){
            l <- numeric(nrow(x))
            l[1] <- 1
            l
        }))))
    Ss <- list()
    for(i_group in seq_len(max(group_index))){
        Ss[[i_group]] <- S[group_index == i_group, ]
    }
    RecError <- NULL
    RelChange <- NULL
    list(A = A, Ss = Ss, RecError = RecError, RelChange = RelChange)
}

.checkGroupICA <- function(Xs, J1, J2, algorithm, num.iter, thr, verbose){
    stopifnot(is.list(Xs))
    nr <- lapply(Xs, nrow)
    all.equal(length(unique(nr)), 1)
    stopifnot(is.numeric(J1))
    stopifnot(length(J1) == 1)
    lapply(Xs, function(x){
        stopifnot(min(dim(x)) >= J1)
    })
    if(algorithm == "Calhoun2009"){
        stopifnot(is.numeric(J2))
        stopifnot(length(J2) == 1)
        # Currently, J2 is treated as $L$ in the paper.
        for(i in seq_len(length(Xs))){
            stopifnot(nrow(Xs[[i]]) >= J2)
            stopifnot(ncol(Xs[[i]]) >= J2)
        }
    }
    stopifnot(is.numeric(num.iter))
    stopifnot(num.iter > 0)
    stopifnot(is.numeric(thr))
    stopifnot(is.logical(verbose))
}

.initGroupICA <- function(Xs, min_sample_per_group = 200){
    # Each X is considered to be each group.
    # For Pfister2018, Xs is a list of data separated by groups.
    # Divide each group equally into partitions.
    # min_sample_per_group is the minimum number of samples in each partition.
    n_group <- length(Xs)
    pooled_X <- do.call(rbind, Xs)
    # Generates a pooled group index according to the number of rows in each element of Xs.
    pooled_grouping <- cumsum(as.numeric(unlist(sapply(
        Xs, function(x){
            s <- numeric(nrow(x)) * 0
            s[1] <- 1
            s
            }))))
    # Divide each group into subgroups with at least min_sample_per_group elements.
    pooled_partition <- c()
    for(i_group in 1:n_group){
        current_X <- Xs[[i_group]]
        current_n <- nrow(current_X)
        min_sample_per_group <- min(floor(current_n / 2), min_sample_per_group)
        pooled_partition <- c(
            pooled_partition,
            .generate_subgroups(current_n, min_sample_per_group))
    }
    list(g = pooled_grouping, Pg = pooled_partition, X = pooled_X, M = list())
}

.eachidx <- function(J1, l, x){
    out <- 1:(J1 * l)
    start <- seq(from = 1, to = J1 * l, by = J1)[x]
    end <- seq(from = J1, to = J1 * l, by = J1)[x]
    out[start:end]
}

.ApproximateJointDiagonalizer <- function(M){
    t(jointDiag::uwedge(M)$B)
}

# Generate subgroups.
# Each subgroup has at least min_sample_per_subgroup samples.
# For example, if n = 9 and min_sample_per_subgroup = 2,
# the result will be 1, 1, 2, 2, 3, 3, 4, 4, 4.
.generate_subgroups <- function(n, min_sample_per_subgroup){
    min_sample_per_subgroup <- as.integer(min_sample_per_subgroup)
    if(n < min_sample_per_subgroup){
        stop("n must be greater than min_sample_per_subgroup")
    }
    # If n is not divisible by min_sample_per_subgroup,
    # each subgroup will have at least min_sample_per_subgroup samples.
    # For example, if n = 9 and min_sample_per_subgroup = 2,
    # the result will be 1, 1, 2, 2, 3, 3, 4, 4, 4.
    if(n %% min_sample_per_subgroup == 0){
        rep(1:(n / min_sample_per_subgroup), each = min_sample_per_subgroup)
    }else{
        # At first, calculate the number of samples that can be divided
        # by min_sample_per_subgroup.
        n_subgroup <- floor(n / min_sample_per_subgroup)
        # then, calculate the remainder of the number of samples divided by
        n_remain <- n %% min_sample_per_subgroup
        # Assign the number of samples that can be divided by min_sample_per_subgroup.
        res <- rep(1:n_subgroup, each = min_sample_per_subgroup)
        # Assign the remainder of the number of samples to the last subgroup.
        res <- c(res, rep(n_subgroup, n_remain))
        res
    }
}

# Simulation Datasets
d <- 6
X1 <- matrix(runif(20 * d), nrow = 20, ncol = d)
X2 <- matrix(runif(15 * d), nrow = 15, ncol = d)
X3 <- matrix(runif(25 * d), nrow = 25, ncol = d)

Xs <- list(X1 = X1, X2 = X2, X3 = X3)

## Perform GroupICA against Simulation Datasets
J1 <- d
set.seed(12345)
out.pooled <- GroupICA(Xs, J1=J1, algorithm="pooled", ica.algorithm = "FastICA", verbose=TRUE)
set.seed(12345)
out.Calhoun2009 <- GroupICA(Xs, J1=J1, algorithm="Calhoun2009", ica.algorithm = "FastICA", verbose=TRUE)
set.seed(12345)
out.Pfister2018 <- suppressWarnings(GroupICA(Xs, J1=J1, algorithm="Pfister2018", ica.algorithm = "FastICA", verbose=TRUE))

set.seed(12345)
out.pooled_2 <- GroupICA(Xs, J1=J1, algorithm="pooled", ica.algorithm = "JADE", verbose=TRUE)
set.seed(12345)
out.Calhoun2009_2 <- GroupICA(Xs, J1=J1, algorithm="Calhoun2009", ica.algorithm = "JADE", verbose=TRUE)
set.seed(12345)
out.Pfister2018_2 <- suppressWarnings(GroupICA(Xs, J1=J1, algorithm="Pfister2018", ica.algorithm = "JADE", verbose=TRUE))

# Test Hidden Functions
set.seed(12345)
out.pooled_ <- iTensor:::.pooled(Xs, J1, J2=J1,
    ica.algorithm="FastICA",
    num.iter=30, thr=1E-10, verbose=FALSE)
set.seed(12345)
out.Calhoun2009_ <- iTensor:::.Calhoun2009(Xs, J1, J2=J1,
    ica.algorithm="FastICA",
    num.iter=30, thr=1E-10, verbose=FALSE)
set.seed(12345)
out.Pfister2018_ <- suppressWarnings(iTensor:::.Pfister2018(Xs, J1, J2=J1,
    ica.algorithm="FastICA",
    num.iter=30, thr=1E-10, verbose=FALSE))
expect_identical(out.pooled, out.pooled_)
expect_identical(out.Calhoun2009, out.Calhoun2009_)
expect_identical(out.Pfister2018, out.Pfister2018_)
expect_true(is.numeric(iTensor:::.eachidx(J1, 1, 1)))
expect_true(is.numeric(iTensor:::.generate_subgroups(1, 1)))

## Test Output object / type
### Test O-1: Object
expect_identical(is.list(out.pooled), TRUE)
expect_identical(is.list(out.Calhoun2009), TRUE)
expect_identical(is.list(out.Pfister2018), TRUE)
expect_identical(is.list(out.pooled_2), TRUE)
expect_identical(is.list(out.Calhoun2009_2), TRUE)
expect_identical(is.list(out.Pfister2018_2), TRUE)

### Test O-2: Object Names
expect_identical(names(out.pooled),
    c("A", "Ss", "RecError", "RelChange"))
expect_identical(names(out.Calhoun2009),
    c("A", "Ss", "RecError", "RelChange"))
expect_identical(names(out.Pfister2018),
    c("A", "Ss", "RecError", "RelChange"))
expect_identical(names(out.pooled_2),
    c("A", "Ss", "RecError", "RelChange"))
expect_identical(names(out.Calhoun2009_2),
    c("A", "Ss", "RecError", "RelChange"))
expect_identical(names(out.Pfister2018_2),
    c("A", "Ss", "RecError", "RelChange"))

### Test O-3: A
expect_equal(
    dim(out.pooled$A),
    c(ncol(Xs[[1]]), J1))
expect_equal(
    dim(out.Calhoun2009$A),
    c(ncol(Xs[[1]]), J1))
expect_equal(
    dim(out.Pfister2018$A),
    c(ncol(Xs[[1]]), J1))
expect_equal(
    dim(out.pooled_2$A),
    c(ncol(Xs[[1]]), J1))
expect_equal(
    dim(out.Calhoun2009_2$A),
    c(ncol(Xs[[1]]), J1))
expect_equal(
    dim(out.Pfister2018_2$A),
    c(ncol(Xs[[1]]), J1))

expect_true(is.matrix(out.pooled$A))
expect_true(is.matrix(out.Calhoun2009$A))
expect_true(is.matrix(out.Pfister2018$A))

expect_true(is.matrix(out.pooled_2$A))
expect_true(is.matrix(out.Calhoun2009_2$A))
expect_true(is.matrix(out.Pfister2018_2$A))

expect_equal(dim(out.pooled$A), c(ncol(Xs[[1]]), J1))
expect_equal(dim(out.Calhoun2009$A), c(ncol(Xs[[1]]), J1))
expect_equal(dim(out.Pfister2018$A), c(ncol(Xs[[1]]), J1))
expect_equal(dim(out.pooled_2$A), c(ncol(Xs[[1]]), J1))
expect_equal(dim(out.Calhoun2009_2$A), c(ncol(Xs[[1]]), J1))
expect_equal(dim(out.Pfister2018_2$A), c(ncol(Xs[[1]]), J1))

### Test O-4: Ss
expect_identical(is.matrix(out.pooled$Ss[[1]]), TRUE)
expect_identical(is.matrix(out.pooled$Ss[[2]]), TRUE)
expect_identical(is.matrix(out.Calhoun2009$Ss[[1]]), TRUE)
expect_identical(is.matrix(out.Calhoun2009$Ss[[2]]), TRUE)
expect_identical(is.matrix(out.Pfister2018$Ss[[1]]), TRUE)
expect_identical(is.matrix(out.Pfister2018$Ss[[2]]), TRUE)
expect_identical(is.matrix(out.pooled_2$Ss[[1]]), TRUE)
expect_identical(is.matrix(out.pooled_2$Ss[[2]]), TRUE)
expect_identical(is.matrix(out.Calhoun2009_2$Ss[[1]]), TRUE)
expect_identical(is.matrix(out.Calhoun2009_2$Ss[[2]]), TRUE)
expect_identical(is.matrix(out.Pfister2018_2$Ss[[1]]), TRUE)
expect_identical(is.matrix(out.Pfister2018_2$Ss[[2]]), TRUE)

expect_equal(dim(out.pooled$Ss[[1]]), c(nrow(Xs[[1]]), J1))
expect_equal(dim(out.pooled$Ss[[2]]), c(nrow(Xs[[2]]), J1))
expect_equal(dim(out.Calhoun2009$Ss[[1]]), c(nrow(Xs[[1]]), J1))
expect_equal(dim(out.Calhoun2009$Ss[[2]]), c(nrow(Xs[[2]]), J1))
expect_equal(dim(out.Pfister2018$Ss[[1]]), c(nrow(Xs[[1]]), J1))
expect_equal(dim(out.Pfister2018$Ss[[2]]), c(nrow(Xs[[2]]), J1))
expect_equal(dim(out.pooled_2$Ss[[1]]), c(nrow(Xs[[1]]), J1))
expect_equal(dim(out.pooled_2$Ss[[2]]), c(nrow(Xs[[2]]), J1))
expect_equal(dim(out.Calhoun2009_2$Ss[[1]]), c(nrow(Xs[[1]]), J1))
expect_equal(dim(out.Calhoun2009_2$Ss[[2]]), c(nrow(Xs[[2]]), J1))
expect_equal(dim(out.Pfister2018_2$Ss[[1]]), c(nrow(Xs[[1]]), J1))
expect_equal(dim(out.Pfister2018_2$Ss[[2]]), c(nrow(Xs[[2]]), J1))

### Test O-5: RecError
expect_identical(is.null(out.pooled$RecError), TRUE)
expect_identical(is.null(out.Calhoun2009$RecError), TRUE)
expect_identical(is.null(out.Pfister2018$RecError), TRUE)
expect_identical(is.null(out.pooled_2$RecError), FALSE)
expect_identical(is.null(out.Calhoun2009_2$RecError), TRUE)
expect_identical(is.null(out.Pfister2018_2$RecError), TRUE)

# JADE does not return relative change
expect_identical(is.null(out.pooled$RelChange), TRUE)
expect_identical(is.null(out.Calhoun2009$RelChange), TRUE)
expect_identical(is.null(out.Pfister2018$RelChange), TRUE)
expect_identical(is.null(out.pooled_2$RelChange), TRUE)
expect_identical(is.null(out.Calhoun2009_2$RelChange), TRUE)
expect_identical(is.null(out.Pfister2018_2$RelChange), TRUE)

## Test Error
### Test E-1: X
expect_error(GroupICA(X, J1=J1))

### Test E-2: J1
expect_error(GroupICA(Xs, J1="5"))
expect_error(GroupICA(Xs, J1=c(2,4)))
expect_error(GroupICA(Xs, J1=10^10))

### Test E-3: J2
expect_error(GroupICA(Xs, J2="5"))
expect_error(GroupICA(Xs, J2=c(2,4)))
expect_error(GroupICA(Xs, J2=10^10))

### Test E-4: algorithm
expect_error(GroupICA(Xs, algorithm="poooled"))

### Test E-5: ica.algorithm
expect_error(GroupICA(Xs, ica.algorithm="JAAE"))

### Test E-6: num.iter
expect_error(GroupICA(Xs, J=J, num_iter="100"))
expect_error(GroupICA(Xs, J=J, num_iter=-1))

### Test E-7: thr
expect_error(GroupICA(X, J=J, thr="0.1"))

### Test E-8: verbose
expect_error(GroupICA(X, J=J, verbose="verbose"))

#' CorrIndex
#'
#' Calculate the CorrIndex of the cross-correlation matrix of S_true and estimated S.
#' The closer the value is to 0, the closer estimated S is to S_true.
#' @param cross_correlation_matrix Cross-correlation matrix
#' @return CorrIndex, which means the closeness between S and S_true , is returned.
#' @examples
#' S_true <- matrix(runif(5*5), nrow=5, ncol=5)
#' S <- matrix(runif(5*5), nrow=5, ncol=5)
#' CorrIndex(cor(S_true, S))
#' @export
CorrIndex <- function(cross_correlation_matrix){
    # 1. Calculate the absolute value of the elements.
    abs_cross_correlation_matrix <- abs(cross_correlation_matrix)
    # 2. Calculate the maximum value for each row.
    max_row <- apply(abs_cross_correlation_matrix, 1, max)
    # 3. Calculate the maximum value for each column.
    max_col <- apply(abs_cross_correlation_matrix, 2, max)
    # 4. Divide by the maximum value for each row.
    row_divided <- abs_cross_correlation_matrix / max_row
    # 5. Divide by the maximum value for each column.
    col_divided <- abs_cross_correlation_matrix / max_col
    # 6. Calculate the sum for each row.
    row_sum <- apply(row_divided, 1, sum)
    # 7. Calculate the sum for each column.
    col_sum <- apply(col_divided, 2, sum)
    # 8. Subtract 1 from the sum for each row.
    row_subtracted <- row_sum - 1
    # 9. Subtract 1 from the sum for each column.
    col_subtracted <- col_sum - 1
    # 10. Calculate the sum of the difference between the sum for each row and 1 and the sum for each column and 1.
    sum(row_subtracted) + sum(col_subtracted)
}

#' Independent Component Analysis (Classic Methods)
#'
#' The input data is assumed to be a matrix.
#' ICA decomposes the matrix and extract the components that
#' are statistically independent each other.
#' @param X A matrix
#' @param J Rank parameter to decompose
#' @param algorithm The decomposition algorithm (Default: "FastICA")
#' @param num.iter The number of iteration
#' @param thr The threshold to terminate the iteration (Default: 1E-10)
#' @param nonlinear_func The function used in FastICA (Default: "tanh")
#' @param learning_rate The learning rate used in InfoMax or ExtInfoMax
#' @param verbose Verbose option
#' @return A list containing the result of the decomposition
#' @examples
#' X <- matrix(runif(100*200), nrow=100, ncol=200)
#' J <- 5
#' out.FastICA <- ICA(X, J=J, algorithm="FastICA")
#' out.InfoMax <- ICA(X, J=J, algorithm="InfoMax")
#' out.ExtInfoMax <- ICA(X, J=J, algorithm="ExtInfoMax")
#' @importFrom MASS ginv
#' @importFrom stats cov
#' @export
ICA <- function(X, J,
    algorithm=c("FastICA", "InfoMax", "ExtInfoMax"),
    num.iter=100, thr=1E-10,
    nonlinear_func=c("tanh", "exp", "kurtosis"),
    learning_rate=1.0, verbose=FALSE){
    ######################################
    # Argument Check
    ######################################
    .checkICA(X, J, num.iter, thr, learning_rate, verbose)
    algorithm <- match.arg(algorithm)
    nonlinear_func <- match.arg(nonlinear_func)
    ######################################
    # Initialization (e.g. Whitening)
    ######################################
    int <- .initICA(X, J, thr=thr, algorithm=algorithm)
    X <- int$X
    W <- int$W
    WhiteningMatrix <- int$WhiteningMatrix
    S <- int$S
    WChange <- int$WChange
    update_func <- int$update_func
    ######################################
    # Iteration
    ######################################
    iter <- 1
    while ((WChange[iter] > thr) && (iter <= num.iter)){
        # Update W
        WNew <- update_func(X, W, S, J, nonlinear_func, learning_rate)
        # After Update
        S <- X %*% WNew
        A <- ginv(WhiteningMatrix %*% WNew)
        X_bar <- S %*% A
        iter <- iter + 1
        # The maximum value of the change in each component
        # of W is defined as the amount of change.
        WChange[iter] <- max(abs(WNew - W))
        W <- WNew
        # Verbose
        if (verbose){
            cat(paste0(iter, " / ", num.iter,
                "max(abs(WNew - W)) = ",
                WChange[iter], "\n"))
        }
    }
    # Output
    names(WChange) <- c("offset", seq_len(iter - 1))
    list(
        A=A, S=S, J=J, algorithm=algorithm, num.iter=num.iter,
        thr=thr, verbose=verbose, WChange=WChange)
}

.checkICA <- function(X, J, num.iter, thr, learning_rate, verbose){
    stopifnot(is.matrix(X))
    stopifnot(is.numeric(J))
    stopifnot(length(J) == 1)
    stopifnot(min(dim(X)) >= J)
    stopifnot(is.numeric(num.iter))
    stopifnot(num.iter >= 0)
    stopifnot(is.numeric(thr))
    stopifnot(is.logical(verbose))
    stopifnot(is.numeric(learning_rate))
}

.initICA <- function(X, J, algorithm, thr = 0.1){
    WhiteningResult <- .whitening(X, J)
    X <- WhiteningResult$XWhiten
    J <- WhiteningResult$JNew
    WhiteningMatrix <- WhiteningResult$WhiteningMatrix
    # initialize recovering matrix
    W <- diag(J)
    # If the initial value of the restoration matrix is not a unit matrix,
    # the following calculation makes sense.
    S <- X %*% W
    X_bar <- S %*% ginv(W)
    # Relative Change
    WChange <- thr * 2
    # select update algorithm
    update_func <- .update[[algorithm]]
    list(X=X, W=W, S=S, WChange=WChange,
        WhiteningMatrix=WhiteningMatrix, update_func=update_func)
}

.whitening <- function(X, J){
    # apply centering
    XMeans <- colMeans(X)
    XCentered <- scale(X, scale=FALSE)
    # compute covariance matrix
    Sigma <- cov(XCentered)
    # compute eigenvectors
    EigenSigma <- eigen(Sigma)
    # Eigenvalues of a covariance matrix are theoretically non-negative.
    # Use abs() to handle small negative values from numerical error.
    absEigenValues <- abs(EigenSigma$values)
    NumNonzeroEigenvalues <- sum(absEigenValues > 0)
    if (NumNonzeroEigenvalues < J){
        J <- NumNonzeroEigenvalues
    }
    LambdaSqrtInv <- diag(1 / sqrt(absEigenValues[1:J]))
    WhiteningMatrix <- EigenSigma$vectors[, 1:J] %*% t(LambdaSqrtInv)
    XWhiten <- XCentered %*% WhiteningMatrix
    # Output
    return(list(XWhiten=XWhiten, JNew=J,
        WhiteningMatrix=WhiteningMatrix, XMeans=XMeans))
}

.orthgonalize <- function(W){
    WNewSvd <- svd(W)
    tcrossprod(WNewSvd$u, WNewSvd$v)
}

# learning_rate is ignored
.FastICA <- function(X, W, S, J, nonlinear_func, learning_rate){
    # Setting
    n <- nrow(X)
    J <- ncol(X)
    # In order to conform to the notation of the paper,
    # W and X is treated as transposed.
    W <- t(W)  # J * J
    X <- t(X)  # J * n
    # Select Non-linear Function
    tmp <- .nonlinear_transform_function_fastica(nonlinear_func)
    g <- tmp$g
    gPrime <- tmp$gPrime
    #########################################
    # Symmetric FastICA
    #########################################
    WX <- W %*% X                                # J * n
    gWX <- g(WX)                                 # J * n
    gPrimeWX <- gPrime(WX)                       # J * n
    EgPrimeWX <- apply(gPrimeWX, 1, FUN = mean)  # J
    EgPrimeWXW <- W %*% diag(EgPrimeWX)          # J * J
    XgWX <- gWX %*% t(X) / n                     # J * J
    WNew <- XgWX - EgPrimeWXW                    # J * J
    # Orthogonalization
    WNew <- .orthgonalize(WNew)
    # Update
    W <- WNew
    # Output
    return(t(W))
}

.nonlinear_transform_function_fastica <- function(nonlinear_func){
    if (nonlinear_func == "tanh"){
        alpha <- 1.0
        g <- function(x){ tanh(alpha* x) }
        gPrime <- function(x){alpha*(1 - tanh(alpha*x)^2)}
    } else if (nonlinear_func == "exp"){
        g <- function(x){x * exp(-x*2/2)}
        gPrime <- function(x){(1-x^2) * exp(-x^2/2)}
    } else if (nonlinear_func == "kurtosis"){
        g <- function(x){x^3}
        gPrime <- function(x){3*x^2}
    }
    list(g=g, gPrime=gPrime)
}

# nonlinear_func is ignored
.InfoMax <- function(X, W, S, J, nonlinear_func, learning_rate){
    ####################################################
    # Gradient method using inverse matrix
    ####################################################
    U <- X %*% W
    Y <- .nonlinear_transform_function_infomax(U)
    delta_W_without_rate <- .dydx_div_dy2dxdw(X, Y, W, nrow(X))
    delta_W <- learning_rate * delta_W_without_rate
    # if you use bias term, include these.
    # delta_W0_without_rate <- dydx_div_dy2dxdw0(X, Y, W)
    # delta_W0 <- learning_rate * delta_W0_without_rate
    # Update
    new_W <- W + delta_W
    # Orthgonalize
    new_W <- .orthgonalize(new_W)
    # Output
    return(new_W)
}

.nonlinear_transform_function_infomax <- function(u){1 / (1 + exp(-u))}

.dydx_div_dy2dxdw <- function(X, Y, W, num_observation){
    return(ginv(t(W)) + crossprod(X, (1-2*Y)) / num_observation)
}

# nonlinear_func is ignored
.ExtInfoMax <- function(X, W, S, J, nonlinear_func, learning_rate){
    ####################################################
    # Gradient method using Natural Gradient
    # It doesn't use inverse matrix.
    ####################################################
    U <- X %*% W
    # Automatically determines whether it is a sub-Gaussian or a super-Gaussian.
    K <- diag(sign(colMeans(.sech(U)^2)*colMeans(U^2) - colMeans(tanh(U)*U)))
    delta_W <- learning_rate * .delta_W_without_learning_rate(U, W, signs=K)
    new_W <- W + delta_W
    # Orthgonalize
    new_W <- .orthgonalize(new_W)
    # Output
    return(new_W)
}

# This is a nonlinear function for both sub-Gaussian and super-Gaussian.
.delta_W_without_learning_rate <- function(U, W, signs){
    return((1-signs %*% crossprod(tanh(U), U) - crossprod(U, U)) %*% W)
}

.sech <- function(x){ 1 / cosh(x) }

.update <- list(
    FastICA=.FastICA,
    InfoMax=.InfoMax,
    ExtInfoMax=.ExtInfoMax)

# Simulation Dataset
X <- matrix(runif(50 * 10), nrow = 50, ncol = 10)

## Perform ICA against Simulation Dataset
J <- 5
out.FastICA <- ICA(X, J=J, algorithm="FastICA")
out.InfoMax <- ICA(X, J=J, algorithm="InfoMax")
out.ExtInfoMax <- ICA(X, J=J, algorithm="ExtInfoMax")

## Test Hidden Functions
expect_equal(length(iTensor:::.nonlinear_transform_function_fastica("tanh")), 2)
expect_true(is.list(iTensor:::.whitening(X, J)))

## Test Output object / type
### Test O-1: Object
expect_identical(is.list(out.FastICA), TRUE)
expect_identical(is.list(out.InfoMax), TRUE)
expect_identical(is.list(out.ExtInfoMax), TRUE)

### Test O-2: Object Names
expect_identical(names(out.FastICA),
    c("A", "S", "J", "algorithm", "num.iter", "thr", "verbose", "WChange"))
expect_identical(names(out.InfoMax),
    c("A", "S", "J", "algorithm", "num.iter", "thr", "verbose", "WChange"))
expect_identical(names(out.ExtInfoMax),
    c("A", "S", "J", "algorithm", "num.iter", "thr", "verbose", "WChange"))

### Test 0-3: A
expect_identical(is.matrix(out.FastICA$A), TRUE)
expect_identical(is.matrix(out.InfoMax$A), TRUE)
expect_identical(is.matrix(out.ExtInfoMax$A), TRUE)

expect_equal(dim(out.FastICA$A), c(J, ncol(X)))
expect_equal(dim(out.InfoMax$A), c(J, ncol(X)))
expect_equal(dim(out.ExtInfoMax$A), c(J, ncol(X)))

### Test 0-4: S
expect_identical(is.matrix(out.FastICA$S), TRUE)
expect_identical(is.matrix(out.InfoMax$S), TRUE)
expect_identical(is.matrix(out.ExtInfoMax$S), TRUE)

expect_equal(dim(out.FastICA$S), c(nrow(X), J))
expect_equal(dim(out.InfoMax$S), c(nrow(X), J))
expect_equal(dim(out.ExtInfoMax$S), c(nrow(X), J))

### Test 0-5: J
expect_identical(out.FastICA$J, J)
expect_identical(out.InfoMax$J, J)
expect_identical(out.ExtInfoMax$J, J)

### Test 0-6: algorithm
expect_identical(out.FastICA$algorithm, "FastICA")
expect_identical(out.InfoMax$algorithm, "InfoMax")
expect_identical(out.ExtInfoMax$algorithm, "ExtInfoMax")

### Test 0-7: num.iter
expect_identical(out.FastICA$num.iter, formals(ICA)$num.iter)
expect_identical(out.InfoMax$num.iter, formals(ICA)$num.iter)
expect_identical(out.ExtInfoMax$num.iter, formals(ICA)$num.iter)

### Test 0-8: thr
expect_identical(out.FastICA$thr, formals(ICA)$thr)
expect_identical(out.InfoMax$thr, formals(ICA)$thr)
expect_identical(out.ExtInfoMax$thr, formals(ICA)$thr)

### Test 0-9: verbose
expect_identical(out.FastICA$verbose, formals(ICA)$verbose)
expect_identical(out.InfoMax$verbose, formals(ICA)$verbose)
expect_identical(out.ExtInfoMax$verbose, formals(ICA)$verbose)

### Test 0-10: RecError
expect_identical(is.vector(out.FastICA$WChange), TRUE)
expect_identical(is.vector(out.InfoMax$WChange), TRUE)
expect_identical(is.vector(out.ExtInfoMax$WChange), TRUE)

## Test Error
### Test E-1: X
expect_error(ICA(as.data.frame(X), J=J))

### Test E-2: J
expect_error(ICA(X, J="5"))
expect_error(ICA(X, J=c(2,4)))
expect_error(ICA(X, J=10^10))

### Test E-3: num.iter
expect_error(ICA(X, J=J, num.iter="100"))
expect_error(ICA(X, J=J, num.iter=-1))

### Test E-4: thr
expect_error(ICA(X, J=J, thr="0.1"))

### Test E-5: verbose
expect_error(ICA(X, J=J, verbose="verbose"))

## Test Orthgonalization
orthgonalized <- iTensor:::.orthgonalize(
    matrix(c(8, 1, 9, 8, 1, 4, 6, 0, 5), 3, 3))
expect_equal((orthgonalized[1, ] %*% orthgonalized[1, ])[1], 1, tolerance=1e-8)
expect_equal((orthgonalized[1, ] %*% orthgonalized[2, ])[1], 0, tolerance=1e-8)
expect_equal((orthgonalized[1, ] %*% orthgonalized[3, ])[1], 0, tolerance=1e-8)
expect_equal((orthgonalized[2, ] %*% orthgonalized[2, ])[1], 1, tolerance=1e-8)
expect_equal((orthgonalized[2, ] %*% orthgonalized[3, ])[1], 0, tolerance=1e-8)

# Simulation Datasets
X <- array(runif(10*20), dim=c(10,20))
Y <- array(runif(15*20), dim=c(15,20))

## Perform MICA against Simulation Datasets
J <- 20
out <- MICA(X, Y, J=J, verbose=TRUE)

## Test Output object / type
### Test O-1: Object
expect_identical(is.list(out), TRUE)

### Test O-2: Object Names
expect_identical(
    names(out),
    c("U", "V", "A", "B", "J", "eta", "num.iter", "verbose", "ABChange"))

### Test O-3: U
expect_identical(is.matrix(out$U), TRUE)
expect_equal(dim(out$U), c(nrow(X), J))

### Test O-4: V
expect_identical(is.matrix(out$V), TRUE)
expect_equal(dim(out$V), c(nrow(Y), J))

### Test O-5: A
expect_identical(is.matrix(out$A), TRUE)
expect_equal(dim(out$A), c(ncol(X), J))

### Test O-6: B
expect_identical(is.matrix(out$B), TRUE)
expect_equal(dim(out$B), c(ncol(Y), J))

### Test O-7: J
expect_identical(out$J, J)

### Test 0-8: eta
expect_identical(out$eta, 1000*1e-4)

### Test O-9: verbose
expect_identical(!out$verbose, formals(MICA)$verbose)

### Test O-10: RecError
expect_identical(is.vector(out$ABChange), TRUE)

## Test Orthgonalization
orthgonalized <- iTensor:::.orthgonalize(
     matrix(c(8, 1, 9, 8, 1, 4, 6, 0, 5), 3, 3))
expect_equal((orthgonalized[1, ] %*% orthgonalized[1, ])[1], 1, tolerance=1e-8)
expect_equal((orthgonalized[1, ] %*% orthgonalized[2, ])[1], 0, tolerance=1e-8)
expect_equal((orthgonalized[1, ] %*% orthgonalized[3, ])[1], 0, tolerance=1e-8)
expect_equal((orthgonalized[2, ] %*% orthgonalized[2, ])[1], 1, tolerance=1e-8)
expect_equal((orthgonalized[2, ] %*% orthgonalized[3, ])[1], 0, tolerance=1e-8)

## Test Error
### Test E-1: X
expect_error(MICA(X, J=J))
expect_error(MICA(Y, J=J))
expect_error(MICA(as.data.frame(X), Y, J=J))
expect_error(MICA(X, as.data.frame(Y), J=J))

### Test E-2: J
expect_error(MICA(X, Y, eta="5"), regexp = "J")
expect_error(MICA(X, Y, eta=c(2,4)), regexp = "J")
expect_error(MICA(X, Y, eta=10^10))

### Test E-3: eta
expect_error(MICA(X, Y, J=J, eta="5"), regexp = "is.numeric(eta)*")
expect_error(MICA(X, Y, J=J, eta=c(2,4)), regexp = "length")

### Test E-4: verbose
expect_error(MICA(X, Y, J=J, verbose="verbose"), regexp = "is.logical")

#' Multimodal independent component analysis
#'
#' The input datasets are assumed to be two matrices sharing the column space.
#' MICA decomposes the matrices simutanously
#' and extracts the components that maximizes the mutual information
#' between the components.
#' @param X A matrix sharing the column space with Y (??? x N)
#' @param Y A matrix sharing the column space with X (??? x N)
#' @param J The rank parameter to decompose the matrices
#' @param eta A learning rate parameter of stochastic gradient descent
#' @param verbose Verbose option
#' @param mu A learning rate parameter of stochastic gradient descent
#' @param gamma_ts Weighting factor for dependence on independence
#' @return A list containing the result of the decomposition
#' @examples
#' X <- array(runif(10*20), dim=c(10,20))
#' Y <- array(runif(15*20), dim=c(15,20))
#' J <- 20
#' out <- MICA(X, Y, J=J)
#' @importFrom stats rnorm
#' @export
MICA <- function(X, Y, J, eta=1000*1e-4, verbose=FALSE,
    mu=50.0*1e-4, gamma_ts=1.0){
    ###########
    # gamma_ts example:
    #   gamma_ts <- 0.2 / (c(1:nSample) + 1)  # used in simulation study of MICA paper
    #   gamma_ts <- rep(0, nSample)           # only independence is evaluated.
    #   gamma_ts <- rep(1, nSample)           # only dependence is evaluated.
    ######################################
    # Argument Check
    ######################################
    .checkMICA(X, Y, J, eta, verbose)
    ######################################
    # Initialization (e.g. CCA)
    ######################################
    int <- .initMICA(X, Y, J)
    A <- int$A
    B <- int$B
    X <- int$XScaled
    Y <- int$YScaled
    nSample <- nrow(X)
    N <- J  # Match the notation for MICA paper.
    ABChange <- 1
    # Time varying learning rate for online estimation of statistics
    if (length(mu) < 2){
        # if constant mu is given, convert it to timeseries.
        mu <- rep(mu, nSample)
    }
    # Weighting factor for dependence on independence
    if (length(gamma_ts) < 2){
        # if constant gamma_ts is given, convert it to timeseries.
        gamma_ts <- rep(gamma_ts, nSample)
    }
    # None of the diagonal elements must be 1.
    # because rho_bar_i is computed using the term R[i,i] / (1 - R[i,i]^2).
    R <- matrix(0, ncol = N, nrow = N)
    # initialize \beta_i^{k, l}
    # definition: \beta_i^{k, l} == beta_i_k_l__zero_origin[i, k+1, l+1]
    beta_i_k_l__zero_origin <- array(data = 1, dim = c(N, 5, 5))
    for (i in 1:N){
        for (k in 0:4){
            for (l in 0:4){
                # use joint cumulants of random normal variable
                beta_i_k_l__zero_origin[i, k+1, l+1] <- mean(
                    rnorm(1000)^k * rnorm(1000)^l) - .beta_zero_k_l(k, l)
            }
        }
    }
    ######################################
    # Iteration
    ######################################
    # use online algorithm
    for (iter in 1:nSample){
        u <- as.numeric(A %*% as.numeric(X[iter, ]))  # size = (N, )
        v <- as.numeric(B %*% as.numeric(Y[iter, ]))  # size = (N, )
        r <- rep(0, N)  # size = (N, )
        s <- rep(0, N)  # size = (N, )
        # Make u and v to uncorrelated variables r and s.
        for (i in 1:N){
            rho_i <- R[i, i]
            c_plus_i <- 1 / 2 * ((1 + rho_i)^(-1/2) + (1 - rho_i)^(-1/2))
            c_minus_i <- 1 / 2 * ((1 + rho_i)^(-1/2) - (1 - rho_i)^(-1/2))
            r[i] <- c_plus_i * u[i] +  c_minus_i * v[i]
            s[i] <- c_minus_i * u[i] +  c_plus_i * v[i]
        }
        # Online estimation of statistic beta_i^{k,l}
        for (i in 1:N){
            for (k in 0:4){
                for (l in 0:4){
                    beta_i_k_l__zero_origin[i, k+1, l+1] <-
                        beta_i_k_l__zero_origin[i, k+1, l+1] -
                            mu[iter] * (
                                beta_i_k_l__zero_origin[i, k+1, l+1] -
                                    r[i]^k * s[i]^l + .beta_zero_k_l(k, l))
                }
            }
        }
        # Online estimation of statistic R
        for (i in 1:N){
            for (j in 1:N){
                R[i, j] <- R[i, j] - mu[iter] * (R[i, j] - u[i]*v[j])
            }
        }
        # .gradA() means (-1) * ∂ε_tot/∂A A^TA
        A_new <- A - eta * .gradA(u, v, A, B, N, J, beta_i_k_l__zero_origin, R, r, s, gamma_ts[iter])
        B_new <- B - eta * .gradB(u, v, A, B, N, J, beta_i_k_l__zero_origin, R, r, s, gamma_ts[iter])
        # Orthgonalize
        A_new <- .orthgonalize(A_new)
        B_new <- .orthgonalize(B_new)
        # Record change of A and B
        # ABChange[iter] <- max(max(abs(A_new - A)), max(abs(B_new - B)))
        ABChange[iter] <- (mean(abs(A_new - A)) + mean(abs(B_new - B))) / 2
        # Verbose
        if (verbose){
             cat(paste0(iter,
                " |max(max(abs(A_new - A)), max(abs(B_new - B)))| = ",
                ABChange[iter], "\n"))
        }
        # Update
        A <- A_new
        B <- B_new
    }
    # Estimate original source using last A and B.
    # Note that this method is difficult in an online setting.
    U <- X %*% t(A)
    V <- Y %*% t(B)
    # Output
    list(U=U, V=V, A=A, B=B, J=J, eta=eta, num.iter=iter,
         verbose=verbose, ABChange=ABChange)
}

.checkMICA <- function(X, Y, J, eta, verbose){
    stopifnot(is.matrix(X))
    stopifnot(is.matrix(Y))
    stopifnot(is.numeric(J))
    stopifnot(length(J) == 1)
    stopifnot(is.numeric(eta))
    stopifnot(length(eta) == 1)
    stopifnot(min(ncol(X), ncol(Y)) >= J)
    stopifnot(is.logical(verbose))
}

.initMICA <- function(X, Y, J){
    XScaled <- scale(X)
    YScaled <- scale(Y)
    A <- diag(J)
    B <- diag(J)
    list(A=A, B=B, XScaled=XScaled, YScaled=YScaled)
}

# factorial function
.fact <- function(x){
    ifelse (x != floor(x) | x < 0, NA, ifelse(x == 0 , 1, prod(1:x)))
}

# phi() function are used in Section: 6  in
# Yang, Howard Hua, and Shun-ichi Amari.
# "Adaptive online learning algorithms for blind separation:
# maximum entropy and minimum mutual information." Neural computation 9.7 (1997): 1457-1482.
.phi <- function(y){2 * tanh(y)}
# phi <- function(y){ y^3 }
# phi <- function(y){ 3 / 4 * y^11 + 15 / 4 * y^9 - 14 / 3 * y^7 - 29 / 4 * y^5 + 29 / 4 * y^3 }

.beta_zero_k_l <- function(k, l){
    if ((k == 4) || (l == 4)){
        return(3)
    } else if ((k == 2) && (l == 2)){
        return(1)
    } else {
        return(0)
    }
}

.der_H_beta_i_der_beta_i_k_l <- function(i, k, l, beta_i_k_l__zero_origin){
    if ((k==3) && (l==0)){
        return(- 1 / (2 * .fact(3) ) * 2 * .beta_i_k_l(i, k, l, beta_i_k_l__zero_origin))
    } else if ((k==2) && (l==1)){
        return(- 1 / (2 * .fact(3) ) * 3 * 2 * .beta_i_k_l(i, k, l, beta_i_k_l__zero_origin))
    } else if ((k==1) && (l==2)){
        return(- 1 / (2 * .fact(3) ) * 3 * 2 * .beta_i_k_l(i, k, l, beta_i_k_l__zero_origin))
    } else if ((k==0) && (l==3)){
        return(- 1 / (2 * .fact(3) ) * 2 * .beta_i_k_l(i, k, l, beta_i_k_l__zero_origin))
    } else if ((k==4) && (l==0)){
        return(- 1 / (2 * .fact(4)) * 2 * .beta_i_k_l(i, k, l, beta_i_k_l__zero_origin))
    } else if ((k==3) && (l==1)){
        return(- 1 / (2 * .fact(4)) * 4 * 2 * .beta_i_k_l(i, k, l, beta_i_k_l__zero_origin))
    } else if ((k==2) && (l==2)){
        return(- 1 / (2 * .fact(4)) * 6 * 2 * .beta_i_k_l(i, k, l, beta_i_k_l__zero_origin))
    } else if ((k==1) && (l==3)){
        return(- 1 / (2 * .fact(4)) * 4 * 2 * .beta_i_k_l(i, k, l, beta_i_k_l__zero_origin))
    } else if ((k==0) && (l==4)){
        return(- 1 / (2 * .fact(4)) * 2 * .beta_i_k_l(i, k, l, beta_i_k_l__zero_origin))
    } else {
        stop("given term is not included in this approximation.")
    }
}

.beta_i_k_l <- function(i, k, l, beta_i_k_l__zero_origin){
    beta_i_k_l__zero_origin[i, k+1, l+1]
}

.der_riksil_der_risiT <- function(k, l, r_i, s_i){
    if (k > 0){
        der_riksil_der_ri <- k * r_i^(k-1) * s_i^l
    } else {
        der_riksil_der_ri <- 0
    }
    if (l > 0){
        der_riksil_der_si <- l * s_i^(l-1) * r_i^k
    } else {
        der_riksil_der_si <- 0
    }
    return(c(der_riksil_der_ri, der_riksil_der_si))
}

.gradA <- function(u, v, A, B, N, J, beta_i_k_l__zero_origin, R, r, s, gamma_t){

    Phi <- rep(0, N)      # size = (N, )
    Xi <- rep(0, N)       # size = (N, )
    Psi <- rep(0, N)      # size = (N, )
    Rho_bar <- rep(0, J)  # size = (N, )

    for (i in 1:N){

        rho_i <- R[i, i]
        c_plus_i <- 1 / 2 * ((1 + rho_i)^(-1/2) + (1 - rho_i)^(-1/2))
        c_minus_i <- 1 / 2 * ((1 + rho_i)^(-1/2) - (1 - rho_i)^(-1/2))

        f_i_g_i <- c(0, 0)
        for (k in 0:5){
            for (l in 0:5){
                # Only all subscript sets of k and l contained in equation (12) are considered.
                if ((k + l <= 2) || (k + l > 4)) next
                f_i_g_i <- f_i_g_i +
                    .der_H_beta_i_der_beta_i_k_l(i, k, l, beta_i_k_l__zero_origin) *
                    .der_riksil_der_risiT(k, l, r[i], s[i])
            }
        }
        f_i <- as.numeric(f_i_g_i[1])
        g_i <- as.numeric(f_i_g_i[2])

        xi_i <- c_plus_i * f_i + c_minus_i * g_i

        d_plus_i <- - 1 / 4 * ((1 + rho_i)^(-3/2) + (1 - rho_i)^(-3/2))
        d_minus_i <- - 1 / 4 * ((1 + rho_i)^(-3/2) - (1 - rho_i)^(-3/2))

        psi_i <- (d_plus_i * u[i] + d_minus_i * v[i]) * f_i + (d_minus_i * u[i] + d_plus_i * v[i]) * g_i
        rho_bar_i <- rho_i / (1-rho_i^2)
        # store them.
        Phi[i] <- .phi(u[i])
        Xi[i] <- xi_i
        Psi[i] <- psi_i
        Rho_bar[i] <- rho_bar_i
    }
    der_A_der_t_div_minus_eta_t <-
        (diag(N) + (1 - gamma_t) * Phi %*% t(u) - gamma_t * Xi %*% t(u) +
             gamma_t * diag(Rho_bar - Psi) %*% t(R)) %*% A
    return(- der_A_der_t_div_minus_eta_t)
}

.gradB <- function(u, v, A, B, N, J, beta_i_k_l__zero_origin, R, r, s, gamma_t){
    .gradA(v, u, B, A, N, J,
        aperm(beta_i_k_l__zero_origin,
            c(1,3,2)), t(R), s, r, gamma_t)
}

#' Independent Component Analysis (Modern Methods)
#'
#' The input data is assumed to be a matrix.
#' ICA decomposes the matrix and extract the components that
#' are statistically independent each other.
#' @param X A matrix
#' @param J Rank parameter to decompose
#' @param algorithm The decomposition algorithm (Default: "JADE")
#' @param num.iter The number of iteration
#' @param thr The threshold to terminate the iteration (Default: 1E-10)
#' @param r_list List of r-th order cumulants used in SIMBEC (Default: NULL)
#' @param omega_for_each_r Weight vector of r_list used in SIMBEC (Default: NULL)
#' @param a_r_for_each_r Parameter vector to specify the shape of partial activation function in SIMBEC (Default: NULL)
#' @param tau_list List of lags to consider the auto-correlation used in AMUSE and SOBI (Default: NULL)
#' @param num_bins Number of bins for histgram in ProDenICA (Default: NULL)
#' @param alpha Learning rate used for gradient descent in RICA (Default: NULL)
#' @param num_epoch Number of epoch used for gradient descent in RICA (Default: NULL)
#' @param verbose Verbose option
#' @return A list containing the result of the decomposition
#' @examples
#' ICA2
#' @importFrom MASS ginv
#' @importFrom graphics hist
#' @importFrom stats cov runif dnorm predict var sd
#' @importFrom mgcv gam
#' @importFrom geigen geigen
#' @importFrom einsum einsum
#' @export
ICA2 <- function(
    X, J,
    algorithm = c(
        "JADE", "AuxICA1", "AuxICA2", "IPCA", "SIMBEC", "AMUSE",
        "SOBI", "FOBI", "ProDenICA", "RICA"),
    num.iter = NULL, thr = 1E-10,
    r_list = NULL,
    omega_for_each_r = NULL,
    a_r_for_each_r = NULL,
    tau_list = NULL,
    num_bins = NULL,
    alpha = NULL,
    num_epoch = NULL,
    verbose = FALSE){
    ######################################
    # Argument Check
    ######################################
    algorithm <- match.arg(algorithm)
    .checkICA2(X, J, num.iter, thr, r_list, omega_for_each_r,
        a_r_for_each_r, tau_list, num_bins,
        alpha, num_epoch, verbose)
    ######################################
    # Iterative or Non-Iterative
    ######################################
    args <- .initICA2_1(X, J, algorithm, num.iter, thr, r_list,
        omega_for_each_r, a_r_for_each_r, tau_list,
        num_bins, alpha, num_epoch, verbose)
    if(algorithm %in% c("JADE", "IPCA", "SIMBEC", "AMUSE", "SOBI", "FOBI")){
        return(do.call(.non_iterative_ICA, args))
    }else if(algorithm %in% c("AuxICA1", "AuxICA2", "ProDenICA", "RICA")){
        return(do.call(.iterative_ICA, args))
    }
}

.non_iterative_ICA <- function(
        X, J,
        algorithm = c("JADE", "IPCA", "SIMBEC", "AMUSE", "SOBI", "FOBI"),
        ...){
    # Argument Check
    algorithm <- match.arg(algorithm)
    args <- list(...)
    verbose <- args[["verbose"]]
    # Performs preprocessing according to the algorithm.
    args[["X"]] <- .preprocess_non_iterative_ICA_input(X = X, J = J, algorithm = algorithm)
    args[["J"]] <- J
    # Runs the algorithm.
    ica_result <- do.call(.ica_function_list[[algorithm]], args)
    # Adds a common attribute to the output that does not depend on the algorithm.
    ica_result[["algorithm"]] <- algorithm
    ica_result[["verbose"]] <- verbose
    ica_result
}

.preprocess_non_iterative_ICA_input <- function(X, J, algorithm){
    # IPCA and JADE differ from the other algorithms in how X is initialized.
    if(algorithm == "IPCA"){
        # IPCA does not apply preprocessing to X.
    }else if(algorithm == "JADE"){
        # For JADE, no scaling is applied to X.
        if(nrow(X) < ncol(X)){
            stop("The number of samples must be larger than the number of dimensions.")
        }
        # Initialize X.
        X_initialized <- .init_X(X, J, scale_X = FALSE)
        X <- X_initialized$X
    }else{
        # For non-IPCA, nrow (X)> = ncol (X).
        if(nrow(X) < ncol(X)){
            stop("The number of samples must be larger than the number of dimensions.")
        }
        # Initialize X.
        X_initialized <- .init_X(X, J)
        X <- X_initialized$X
    }
    X
}

.iterative_ICA <- function(
        X, J,
        algorithm = c("AuxICA1", "AuxICA2", "ProDenICA", "RICA"),
        num.iter = 30, thr = 1E-10, verbose = FALSE, ...){
    ######################################
    # Argument Check
    ######################################
    algorithm <- match.arg(algorithm)
    args <- list(...)
    args[["X"]] <- X
    args[["J"]] <- J
    args[["algorithm"]] <- algorithm
    args[["num.iter"]] <- num.iter
    args[["thr"]] <- thr
    args[["verbose"]] <- verbose
    ######################################
    # Initialization (e.g. Whiteniing)
    ######################################
    if(algorithm == "RICA"){
        allow_overcomplete <- TRUE
    }else{
        allow_overcomplete <- FALSE
    }
    int <- .initICA2_2(X, J, thr, allow_overcomplete)
    X <- int$X
    A <- int$A
    W <- int$W
    S <- int$S
    RecError <- int$RecError
    RelChange <- int$RelChange
    ######################################
    # Iteration
    ######################################
    iter <- 1
    # TODO: Replace the following line with the next line.
    # while ((RelChange[iter] > thr) && (iter <= num.iter)){
    while (iter <= num.iter){
        # Runs the algorithm.
        args[["X"]] <- X
        args[["W"]] <- W
        args[["J"]] <- J
        ica_result <- do.call(.ica_function_list[[algorithm]], args)
        W <- ica_result$W
        # After Update
        A <- ginv(W)
        S <- X %*% W
        X_bar <- S %*% A
        iter <- iter + 1
        RecError[iter] <- .recError(X, X_bar)
        RelChange[iter] <- .relChange(iter, RecError)
        # Verbose
        if(verbose){
            cat(paste0(
                iter, " / ", num.iter,
                " |Previous Error - Error| / Error = ",
                RelChange[iter], "\n"
            ))
        }
    }
    # Output
    names(RecError) <- c("offset", 1:(iter - 1))
    names(RelChange) <- c("offset", 1:(iter - 1))
    list(
        A = A, S = S, J = J, W = W, X = X, algorithm = algorithm, num.iter = num.iter,
        thr = thr, verbose = verbose, RecError = RecError, RelChange = RelChange
    )
}

.JADE <- function(X, W = NULL, J = NULL, num.iter = 500, verbose = FALSE, ...){
    num_dimensions <- ncol(X)
    # Estimates the cumulant tensor.
    # NOTE: It is a good idea to estimate the covariance matrix
    #       in advance if you want to speed up the following process.
    Q <- array(0.0, dim = rep(num_dimensions, 4))
    for(i in 1:num_dimensions){
        for(j in 1:num_dimensions){
            for(k in 1:num_dimensions){
                for(l in 1:num_dimensions){
                    Q[i, j, k, l] <- mean(X[, i] * X[, j] * X[, k] * X[, l]) -
                        mean(X[, i] * X[, j]) * mean(X[, k] * X[, l]) -
                        mean(X[, i] * X[, k]) * mean(X[, j] * X[, l]) -
                        mean(X[, i] * X[, l]) * mean(X[, j] * X[, k])
                }
            }
        }
    }
    Q_unfolded <- matrix(as.numeric(Q), nrow = num_dimensions^2, ncol = num_dimensions^2)
    eigen_Q <- eigen(Q_unfolded)
    # jointly diagonalise the set \hat{N}^e = \{ \hat{lambda}_r \hat{M}_r | 1 \le r \le n}
    n_eigen_matrices_diagonalized <- num_dimensions
    num_components <- num_dimensions
    # Create a list of λ M in array format.
    # The last dimension is the number of the eigenmatrix to be diagonalized.
    lambda_M_list <- array(
        0.0,
        dim = c(num_dimensions, num_dimensions, n_eigen_matrices_diagonalized))
    order_of_abs_eigenvalues <- order(abs(eigen_Q$values), decreasing = TRUE)
    for(i in 1:n_eigen_matrices_diagonalized){
        lambda_M_list[, , i] <- eigen_Q$values[order_of_abs_eigenvalues[i]] *
            matrix(eigen_Q$vectors[, order_of_abs_eigenvalues[i]], num_dimensions, num_dimensions)
    }
    V <- diag(num_components)
    for(i_loop in 1:num.iter){
        for(p in 1:num_components){
            for(q in 1:num_components){
                if(p == q) next
                # Use Cyclic-by-Row Algorithm
                # Reference: Gene H. Golub, Charles F. Van Loan Matrix Computations p.480
                g_r <- matrix(NA, nrow = num_components, ncol = 3)  # An array of num_component * 3.
                for(r in 1:num_components){
                    g_r[r, ] <- c(
                        lambda_M_list[p, p, r] - lambda_M_list[q, q, r],
                        lambda_M_list[p, q, r],
                        lambda_M_list[q, p, r]
                    )
                }
                B <- matrix(c(1, 0, 0, 0, 1, 1, 0, -1i, 1i), nrow = 3, ncol = 3, byrow = TRUE)
                ReGHG <- Re(B %*% t(g_r) %*% g_r %*% t(B))
                ReGHG_eigen <- eigen(ReGHG)
                angles <- ReGHG_eigen$vectors[, 1]
                if(angles[1] < 0) angles <- -angles
                c_ <- sqrt(0.5 + angles[1] / 2)
                s_ <- 0.5 * (angles[2] - 1i * angles[3]) / c_
                # Creates a Givens rotation matrix.
                GivensRot <- .R_ijcs(p, q, c_, s_, num_components)
                V <- V %*% GivensRot
                for(r in 1:num_components){
                    lambda_M_list[, , r] <- t(GivensRot) %*% lambda_M_list[, , r] %*% GivensRot
                }
            }
        }
    }
    A <- V
    un_mixing_matrix <- solve(A)
    S <- X %*% V
    X_bar <- S %*% A
    # Output
    list(
        A = A,
        W = un_mixing_matrix,
        S = S,
        X = X_bar,
        J = J,
        num.iter = 1,
        thr = NULL,
        RecError = .recError(X, X_bar),  # Returns only the last reconstruction error.
        RelChange = NULL
    )
}

# Generate Givens rotation matrix.
.R_ijcs <- function(i, j, c, s, n){
    R <- diag(n)
    R[i, i] <- c
    R[i, j] <- -Conj(s)
    R[j, i] <- s
    R[j, j] <- c
    R
}

.AuxICA1 <- function(X, W = NULL, J = NULL, verbose = FALSE, ...){
    stopifnot(!is.null(W))
    # Within this function, we treat the rows as sensors and the columns as samples,
    # according to the notation in the paper.
    # X_t: J * n
    X_t <- t(X)
    n <- nrow(X)
    # W: J * J
    # Each row of W is a row vector of w_1, w_2, ... w_J.
    W <- t(W)
    G_prime <- function(x){
        # Let G (x) = log(cosh(abs(x))), where G'(x) = tanh(x).
        return(tanh(x))
    }
    for(k in 1:J){
        # Auxiliary variable updates
        w_k <- W[k, ]
        # n * J * J
        V_k_i_list <- array(0.0, dim = c(n, J, J))
        for(i in 1:n){
            x <- X_t[, i]
            # r_k is inner product of w_k and x (scalar)
            r_k <- abs(as.numeric(w_k %*% x))
            # J * J
            V_k_i <- G_prime(r_k) / r_k * outer(x, x)
            V_k_i_list[i, , ] <- V_k_i
        }
        # J * J
        V_k <- apply(V_k_i_list, c(2, 3), mean)
        # (J - 1) * J
        W_k_minus_1 <- t(W[-k, ])
        # J * (J - 1)
        P <- V_k %*% W_k_minus_1
        # Demixing matrix updates
        w_k <- w_k - P %*% solve(t(P) %*% P) %*% t(P) %*% w_k
        w_k <- w_k / sqrt(as.numeric(t(w_k) %*% V_k %*% w_k))
        W[k, ] <- w_k
    }
    # Output
    list(
        A = NULL,
        W = t(W),
        S = NULL,
        X = X,
        J = J,
        num.iter = NULL,
        thr = NULL,
        RecError = NULL,
        RelChange = NULL
    )
}

.AuxICA2 <- function(X, W = NULL, J = NULL, verbose = FALSE, ...){
    # Within this function, we treat the rows as sensors and the columns as samples,
    # according to the notation in the paper.
    # X_t: J * num_sample
    X_t <- t(X)
    num_sample <- nrow(X)
    # W: J * J
    # Each row of W is a row vector of w_1, w_2, ... w_J.
    W <- t(W)
    G_prime <- function(x){
        # Let G(x) = log(cosh(abs(x))), where G'(x) = tanh(abs(x)).
        return(tanh(x))
    }
    for(m in 1:(J - 1)){
        for(n in (m + 1):J){
            # Auxiliary variable updates
            # J * 1
            w_m <- W[m, ]
            # J * 1
            w_n <- W[n, ]
            # n * 2 * 2
            U_m_i_list <- array(0.0, dim = c(num_sample, 2, 2))
            # n * 2 * 2
            U_n_i_list <- array(0.0, dim = c(num_sample, 2, 2))
            for(i in 1:num_sample){
                x <- X_t[, i]
                w_m_h_x <- as.numeric(w_m %*% x)
                w_n_h_x <- as.numeric(w_n %*% x)
                u <- c(w_m_h_x, w_n_h_x)
                r_m <- abs(w_m_h_x)
                r_n <- abs(w_n_h_x)
                # 2 * 2
                U_m_i <- G_prime(r_m) / r_m * outer(u, u)
                U_n_i <- G_prime(r_n) / r_n * outer(u, u)

                U_m_i_list[i, , ] <- U_m_i
                U_n_i_list[i, , ] <- U_n_i
            }
            # J * J
            U_m <- apply(U_m_i_list, c(2, 3), mean)
            U_n <- apply(U_n_i_list, c(2, 3), mean)
            # Demixing matrix updates
            eig <- geigen(U_m, U_n, symmetric = TRUE)
            h_m <- eig$vectors[, 1]
            h_n <- eig$vectors[, 2]
            h_m <- h_m / sqrt(as.numeric(t(h_m) %*% U_m %*% h_m))
            h_n <- h_n / sqrt(as.numeric(t(h_n) %*% U_n %*% h_n))
            w_m_w_n <- cbind(w_m, w_n) %*% cbind(h_m, h_n)
            w_m <- w_m_w_n[, 1]
            w_n <- w_m_w_n[, 2]
            W[m, ] <- w_m
            W[n, ] <- w_n
        }
    }
    # Output
    list(
        A = NULL,
        W = t(W),
        S = NULL,
        X = X,
        J = J,
        num.iter = NULL,
        thr = NULL,
        RecError = NULL,
        RelChange = NULL
    )
}

.IPCA <- function(X, W = NULL, J = NULL, verbose = FALSE, num.iter = 200, ...){
    # IPCA assumes num_columns > num_rows.
    num_components <- J
    X_centered <- scale(X, scale = FALSE) # size: n * p
    X_svd <- svd(X_centered)
    V <- X_svd$v # p * n
    U <- X_svd$u
    V_scaled <- scale(V) # p * n
    V_t <- t(V_scaled)[1:num_components, ] # n * p -> num_components * p
    transposed_Vt_whitening_result <- .whitening2(t(V_t))
    # initial_W: num_components * num_components
    initial_W <- diag(num_components)
    # W_: num_components * num_components
    W_ <- initial_W
    for(i_iteration in 1:num.iter){
        W_ <- .FastICA2(
            transposed_Vt_whitening_result$XWhiten,
            W = W_, S = NULL, J = num_components)
    }
    W_ <- solve(W_)
    # S: num_components * p
    S <- W_ %*% V_t
    kurtosis <- function(x){
        X_scaled <- scale(x)
        mean(X_scaled^4) - 3
    }
    S_kurtosis <- apply(S, MARGIN = 1, FUN = kurtosis)
    order_S_kurtosis <- order(S_kurtosis, decreasing = TRUE)
    S <- S[order_S_kurtosis, ]
    U_tilde <- X_centered %*% t(S) # n * num_components
    X_bar <- U_tilde %*% ginv(t(S))
    # Output
    list(
        A = ginv(t(S)),
        W = t(S),
        S = U_tilde,
        X = X_bar,
        J = J,
        num.iter = 1,
        thr = NULL,
        RecError = .recError(X_centered, X_bar),
        RelChange = NULL
    )
}

.SIMBEC <- function(X, W = NULL, J = NULL, verbose = FALSE, num.iter = 1000,
                    r_list = c(4), omega_for_each_r = c(1.0),
                    a_r_for_each_r = c(1.0), ...){
    # Within this function, we treat the rows as sensors and the columns as samples,
    # according to the notation in the paper.
    X_centered <- t(X)
    n_r <- length(r_list)
    E <- J
    N <- nrow(X_centered)
    # U: E * N
    U <- matrix(0.0, nrow = E, ncol = N)
    for(i_component in 1:E){
        U[i_component, i_component] <- 1.0
    }
    for(i_iter in 1:num.iter){
        Y <- U %*% X_centered
        Y_centered <- scale(Y, scale = FALSE)
        # |C^4_{y_i}| ... forth_order_auto_cumulants
        # Reference: https://mathworld.wolfram.com/Cumulant.html
        forth_order_auto_cumulants_of_y <- rowMeans(Y_centered^4) - 3 * apply(Y_centered, 1, var)^2
        # \mu
        mu <- 2 / (3 * max(abs(forth_order_auto_cumulants_of_y)))
        added_term_without_mu <- matrix(0.0, nrow = E, ncol = N)
        for(i_r in 1:n_r){
            # Within this loop, we compute:
            # \omega _{r}( D_{2}^{r}C_{y,x}^{r-1,1}-C_{y,y}^{1,r-1}D_{z}^{r}U^{\left( k\right) }
            r <- r_list[i_r]
            omega_r <- omega_for_each_r[i_r]
            # |C^r_{y_i}| ... r-th_order_auto_cumulants
            # refer: https://mathworld.wolfram.com/Cumulant.html
            if(r == 4){
                r_th_order_auto_cumulants_of_y <- rowMeans(Y_centered^4) - 3 * apply(Y_centered, 1, var)^2
            }else if(r == 3){
                r_th_order_auto_cumulants_of_y <- rowMeans(Y_centered^3)
            }else{
                stop("r must be 3 or 4.")
            }
            # \alpha_r
            alpha_r <- a_r_for_each_r[i_r]
            # \boldmath{D}^r_y
            # Size is E * E
            # It is computed by the elementwise product.
            D_r_y <- diag(sign(r_th_order_auto_cumulants_of_y) * abs(r_th_order_auto_cumulants_of_y)^(alpha_r - 1))
            # \boldmath{C}_{y,x}^{r-1,1}
            # Size is E * N
            # E(xi^3 * xj) - E(xixj)E(xixi) - E(xixj)E(xixi) - E(xixj)E(xixi)
            # = E(xi^3 xj) - 3E(xixj)E(xixi)
            r_th_order_cross_cumulants_of_x_y <- matrix(0.0, nrow = E, ncol = N)
            for(i_row in 1:E){
                for(i_col in 1:N){
                    r_th_order_cross_cumulants_of_x_y[i_row, i_col] <- mean(
                        Y_centered[i_row, ]^(r - 1) * X_centered[i_col, ]
                    ) -
                        3 * mean(Y_centered[i_row, ] * X_centered[i_col, ]) *
                            mean(Y_centered[i_row, ] * Y_centered[i_row, ])
                }
            }
            # \boldmath{C}_{y,y}^{1,r-1}
            # Size is E * E
            # E(xi * xj^3) - E(xixj)E(xjxj) - E(xixj)E(xjxj) - E(xixj)E(xjxj)
            # = E(xi xj^3) - 3E(xixj)E(xjxj)
            r_th_order_cross_cumulants_of_y_y <- matrix(0.0, nrow = E, ncol = E)
            for(i_row in 1:E){
                for(i_col in 1:E){
                    r_th_order_cross_cumulants_of_y_y[i_row, i_col] <- mean(
                        Y_centered[i_row, ] * Y_centered[i_col, ]^(r - 1)
                    ) -
                        3 * mean(Y_centered[i_row, ] * Y_centered[i_col, ]) *
                            mean(Y_centered[i_col, ] * Y_centered[i_col, ])
                }
            }
            added_term_without_mu <- added_term_without_mu +
                omega_r * (
                    D_r_y %*% r_th_order_cross_cumulants_of_x_y -
                        r_th_order_cross_cumulants_of_y_y %*% D_r_y %*% U)
        }
        U <- U + mu * added_term_without_mu
        U <- .orthgonalize(U)
    }
    S <- t(U %*% X_centered)
    A <- ginv(t(U))
    X_bar <- S %*% A
    # Output
    list(
        A = A,
        W = t(U),
        S = S,
        X = X_bar,
        J = J,
        num.iter = 1,
        thr = NULL,
        RecError = .recError(t(X_centered), X_bar), # Returns only the estimate completion error.
        RelChange = NULL
    )
}

# Converts the given matrix to an orthonormal matrix
.orthgonalize <- function(W){
    W_svd <- svd(W)
    tcrossprod(W_svd$u, W_svd$v)
}

.AMUSE <- function(X, W = NULL, J = NULL, verbose = FALSE, tau_list = 1:50, ...){
    # This function treats rows as sensors and columns as samples (time).
    X_centered <- t(X)
    Y <- t(X_centered)
    max_eigen_values <- c()
    for(tau in tau_list){
        n <- nrow(Y)
        R_y_tau <- cov(Y[1:(n - tau), ], Y[(1 + tau):n, ])
        eig_R_y_tau <- eigen((R_y_tau + t(R_y_tau)) / 2)
        max_eigen_values <- c(max_eigen_values, max(eig_R_y_tau$values))
    }
    tau <- which.max(max_eigen_values)
    R_y_tau <- cov(Y[1:(n - tau), ], Y[(1 + tau):n, ])
    eig_R_y_tau <- eigen((R_y_tau + t(R_y_tau)) / 2)
    W <- eig_R_y_tau$vectors
    A <- ginv(W)
    S <- t(X_centered) %*% W
    X_bar <- S %*% A
    # Output
    list(
        A = ginv(W),
        W = W,
        S = S,
        X = X_bar,
        J = J,
        num.iter = 1,
        thr = NULL,
        # Return only the error at last estimate.
        RecError = .recError(t(X_centered), X_bar),
        RelChange = NULL
    )
}

.SOBI <- function(X, W = NULL, J = NULL, verbose = FALSE, num.iter = 100, tau_list = 1:10, ...){
    num_components <- J
    # Create Givens rotation matrix.
    R_ijcs <- function(i, j, c, s, n, use_complex = FALSE){
        R <- diag(n)
        R[i, i] <- c
        if(use_complex){
            R[i, j] <- -Conj(s)
        }else{
            R[i, j] <- -s
        }
        R[j, i] <- s
        R[j, j] <- c
        R
    }
    # This function treats rows as sensors and columns as samples (time).
    X_centered <- t(X)
    n_covariance_matrices <- length(tau_list)
    # Combines the covariance matrices into an array.
    covariance_matrices <- array(
        0.0,
        dim = c(num_components, num_components, n_covariance_matrices)
    )
    Y <- t(X_centered)
    n <- nrow(Y)
    for(i_tau in 1:n_covariance_matrices){
        tau <- tau_list[i_tau]
        R_tau <- cov(Y[1:(n - tau), ], Y[(1 + tau):n, ])
        covariance_matrices[, , i_tau] <- R_tau
    }
    # NOTE: You may be able to change to an iterative approach by tweaking this.
    V <- diag(num_components)
    for(i_loop in 1:num.iter){
        for(p in 1:num_components){
            for(q in 1:num_components){
                if(p == q) next
                # Use Cyclic-by-Row Algorithm
                # See Golub and Loan Matrix Computations p.430.
                # Size is num_components * 3
                g_r <- matrix(NA, nrow = num_components, ncol = 3)
                for(r in 1:num_components){
                    g_r[r, ] <- c(
                        covariance_matrices[p, p, r] - covariance_matrices[q, q, r],
                        covariance_matrices[p, q, r],
                        covariance_matrices[q, p, r]
                    )
                }
                ReGHG <- Re(t(g_r) %*% g_r)
                ReGHG_eigen <- eigen(ReGHG)
                angles <- ReGHG_eigen$vectors[, 1]
                c_ <- sqrt(0.5 + angles[1] / 2)
                use_complex <- FALSE
                if(use_complex){
                    s_ <- 0.5 * (angles[2] - 1i * angles[3]) / c_
                }else{
                    s_ <- 0.5 * angles[2] / c_
                }
                GivensRot <- R_ijcs(p, q, c_, s_, num_components, use_complex = use_complex)
                V <- V %*% GivensRot
                for(r in 1:num_components){
                    covariance_matrices[, , r] <- t(GivensRot) %*% covariance_matrices[, , r] %*% GivensRot
                }
            }
        }
    }
    W <- V
    A <- ginv(V)
    S <- t(X_centered) %*% W
    X_bar <- S %*% A
    # Output
    list(
        A = ginv(W),
        W = W,
        S = S,
        X = X_bar,
        J = J,
        num.iter = 1,
        thr = NULL,
        # Return only the error at last estimate.
        RecError = .recError(t(X_centered), X_bar),
        RelChange = NULL
    )
}

.FOBI <- function(X, W = NULL, J = NULL, verbose = FALSE, ...){
    X_t <- t(X)
    X_norm <- sqrt(colSums(X_t^2))
    X_mul_norm <- t(t(X_t) * X_norm)
    Omega <- X_mul_norm %*% t(X_mul_norm)
    eigen_Omega <- eigen(Omega)
    unmixing_mat <- t(eigen_Omega$vectors)
    W <- unmixing_mat
    A <- ginv(W)
    S <- t(X_t) %*% W
    X_bar <- S %*% A
    # Output
    list(
        A = A,
        W = W,
        S = S,
        X = X_bar,
        J = J,
        num.iter = 1,
        thr = NULL,
        RecError = .recError(t(X_t), X_bar),
        RelChange = NULL
    )
}

.ProDenICA <- function(
        X, W = NULL, J = NULL,
        num.iter = 100, num_bins = 20,
        verbose = FALSE, ...){
    n_sample <- nrow(X)
    for(i_iter in 1:num.iter){
        W <- .orthgonalize(W)
        S <- X %*% W
        Gs <- matrix(NA, nrow = n_sample, ncol = J)
        G_prime_primes <- matrix(NA, nrow = n_sample, ncol = J)
        for(j in 1:J){
            # Estiamte G
            s_vector <- scale(S[, j])
            bins <- seq(min(s_vector), max(s_vector),, num_bins + 1)
            count_on_grid <- hist(s_vector, breaks = bins, plot = FALSE)
            y <- count_on_grid$counts
            grid <- count_on_grid$mids
            gauss <- dnorm(grid)
            gam_model <- gam(
                y ~ s(grid, k = 6), family = "poisson", offset = log(gauss))
            # Estimate G'
            fine_grid <- seq(min(s_vector), max(s_vector), length.out = 1000)
            G_fine_grid <- predict(
                gam_model, newdata = data.frame(grid = fine_grid), type = "response")
            log_G_fine_grid_without_gauss <- log(G_fine_grid / dnorm(fine_grid))

            gam_model_deriv_log_G_fine_grid_without_gauss <- gam(
                y ~ s(fine_grid, k = 6), family = "gaussian",
                data = list(
                    y = diff(log_G_fine_grid_without_gauss) / diff(range(fine_grid)) * length(fine_grid),
                    fine_grid = fine_grid[2:length(fine_grid)]))
            # Estimate G''
            gam_model_deriv_deriv_log_G_fine_grid_without_gauss <- gam(
                y ~ s(fine_grid, k = 6), family = "gaussian",
                data = list(
                    y = diff(diff(log_G_fine_grid_without_gauss)) / diff(range(fine_grid)) * length(fine_grid),
                    fine_grid = fine_grid[3:length(fine_grid)]))
            Gs[, j] <- predict(
                gam_model_deriv_log_G_fine_grid_without_gauss,
                newdata = data.frame(fine_grid = s_vector), type = "response")
            G_prime_primes[, j] <- predict(
                gam_model_deriv_deriv_log_G_fine_grid_without_gauss,
                newdata = data.frame(fine_grid = s_vector), type = "response")
        }
        # Fixed point algorithm
        X_Gs_div_n_sample <- t(X) %*% Gs / n_sample
        W <- X_Gs_div_n_sample - scale(
            W, center = FALSE, scale = 1 / colMeans(G_prime_primes))
    }
    W <- .orthgonalize(W)
    # Output
    list(
        A = NULL,
        W = W,
        S = NULL,
        X = X,
        J = J,
        num.iter = NULL,
        thr = NULL,
        RecError = NULL,
        RelChange = NULL
    )
}

# Solve Rica with SGD.
.RICA <- function(X, W = NULL, J = NULL, verbose = FALSE, alpha = 0.001, num_epoch = 100, ...){
    # NOTE: RICA can also be used for overcomplete condition.
    K_component <- J
    n_input <- J
    # Within this function, we treat rows as sensors and columns as samples (time).
    num_sample <- nrow(X)
    X_whiten <- t(X)
    rica_loss_and_gradient_one_sample <- function(W, x, K_component, n_input){
        # Calculate with the following configuration:
        # 1. Reconstruction error term
        # 1-1. Forward
        # 1-2. Reverse Direction
        # 2. ICA term
        # 2-1. Forward
        # 2-2. Reverse Direction
        ## 1. Reconstruction error term
        ### 1-1. Forward
        y040 <- t(W) %*% W
        y030 <- c(y040 %*% x)
        y020 <- y030 - x
        y010 <- y020^2
        L <- sum(y010)
        ### 1.2. Reverse Direction
        dL_dy010 <- rep(1, n_input)
        dy010_dy020 <- 2 * y020
        dy020_dy030 <- rep(1, n_input)
        dy030_dy040 <- array(0, dim = c(n_input, n_input, n_input))
        for(i in 1:n_input){
            for(j in 1:n_input){
                for(k in 1:n_input){
                    dy030_dy040[i, j, k] <- (i == k) * x[j]
                }
            }
        }
        dy040_dW <- array(0, dim = c(K_component, n_input, n_input, n_input))
        for(i in 1:K_component){
            for(j in 1:n_input){
                for(k in 1:n_input){
                    for(l in 1:n_input){
                        dy040_dW[i, j, k, l] <- (j == k) * W[i, l] + (j == l) * W[i, k]
                    }
                }
            }
        }
        interim <- einsum("k,ijk->ij", dL_dy010 * dy010_dy020 * dy020_dy030, dy030_dy040)
        diff_f_W_estimate <- einsum("kl,ijkl->ij", interim, dy040_dW)
        ## 2. ICA term
        g <- function(x){
            1/ 2 * log(cosh(2 * x))
        }
        diff_g <- function(x){
            tanh(2 * x)
        }
        ### 2.1. Forward
        z020 <- c(W %*% x)
        z010 <- g(z020)
        Lica <- sum(z010)
        ### 2.2. Reverse Direction
        dLica_dz010 <- rep(1, K_component)
        dz010_dz020 <- diff_g(z020)
        dz020_dW <- array(0, dim = c(K_component, n_input, K_component))
        for(i in 1:K_component){
            for(j in 1:n_input){
                for(k in 1:K_component){
                    dz020_dW[i, j, k] <- (i == k) * x[j]
                }
            }
        }
        diff_W_ICA_estimate <- einsum(
            "k,ijk->ij",
            dLica_dz010 * dz010_dz020, dz020_dW
        )
        # Output
        return(list(
            loss = L + Lica,
            grad = diff_f_W_estimate + diff_W_ICA_estimate,
            L = L, Lica = Lica, diff_f_W_estimate = diff_f_W_estimate,
            diff_W_ICA_estimate = diff_W_ICA_estimate
        ))
    }
    W_ <- W
    for(i_epoch in 1:num_epoch){
        # decay learning rate
        alpha <- alpha * 0.99
        sum_loss <- 0
        sum_L <- 0
        sum_Lica <- 0
        accumulated_grad <- matrix(0, nrow = K_component, ncol = n_input)
        for(i_sample in 1:num_sample){
            xi <- X_whiten[, i_sample]
            result <- rica_loss_and_gradient_one_sample(W_, xi, K_component, n_input)
            accumulated_grad <- accumulated_grad + result$grad
            # W_ <- W_ - alpha * result$grad  # TODO: For SGD, use this line.
            sum_loss <- sum_loss + result$loss
            sum_L <- sum_L + result$L
            sum_Lica <- sum_Lica + result$Lica
        }
        W_ <- W_ - alpha * accumulated_grad / num_sample
        if(verbose){
            cat("epoch: ", i_epoch, " loss: ", sum_loss,
                " L: ", sum_L, " Lica: ", sum_Lica,
                " diff_f_W_estimate: ", result$diff_f_W_estimate,
                " diff_W_ICA_estimate: ", result$diff_W_ICA_estimate,
                " W_: ", W_, ", ",
                sep = "\n"
            )
        }
    }
    # Output
    list(
        A = NULL,
        W = t(W_),
        S = NULL,
        X = X,
        J = J,
        num.iter = NULL,
        thr = NULL,
        RecError = NULL,
        RelChange = NULL
    )
}

.checkICA2 <- function(X, J, num.iter, thr, r_list,
    omega_for_each_r, a_r_for_each_r, tau_list,
    num_bins, alpha, num_epoch, verbose){
    stopifnot(is.matrix(X))
    stopifnot(is.numeric(J))
    stopifnot(length(J) == 1)
    stopifnot(min(dim(X)) >= J)
    if(!is.null(num.iter)){
        stopifnot(is.numeric(num.iter))
        stopifnot(num.iter >= 1)
    }
    stopifnot(is.numeric(thr))
    if(!is.null(r_list)){
        stopifnot(is.numeric(r_list))
    }
    if(!is.null(omega_for_each_r)){
        stopifnot(is.numeric(omega_for_each_r))
    }
    if(!is.null(a_r_for_each_r)){
        stopifnot(is.numeric(a_r_for_each_r))
    }
    if(!is.null(tau_list)){
        stopifnot(is.numeric(tau_list))
    }
    if(!is.null(num_bins)){
        stopifnot(is.numeric(num_bins))
    }
    if(!is.null(alpha)){
        stopifnot(is.numeric(alpha))
    }
    if(!is.null(num_epoch)){
        stopifnot(is.numeric(num_epoch))
    }
    stopifnot(is.logical(verbose))
}

.init_X <- function(X, J, scale_X = TRUE){
    # Gets the mean and standard deviation vectors of X.
    XMeans <- colMeans(X)
    XSds <- apply(X, 1, sd)
    # Whitens X.
    WhiteningResult <- .whitening2(X, scale = scale_X)
    X <- WhiteningResult$XWhiten
    WhiteningMatrix <- WhiteningResult$WhiteningMatrix
    # Truncate X according to the number of elements specified.
    # if J <ncol (X), truncate.
    # if ncol (X) <= J, leave as is.
    if(J < ncol(X)){
        X <- X[, 1:J]
    }
    # Output
    list(X = X,
        XMeans = XMeans,
        XSds = XSds,
        WhiteningMatrix = WhiteningMatrix
    )
}

.initICA2_1 <- function(X, J, algorithm, num.iter, thr, r_list,
    omega_for_each_r, a_r_for_each_r, tau_list,
    num_bins, alpha, num_epoch, verbose){
    # Algorithm's default parameters
    if(algorithm == "JADE"){
        num.iter <- if(is.null(num.iter)) 500 else num.iter
    }else if(algorithm == "AuxICA1"){
        num.iter <- if(is.null(num.iter)) 30 else num.iter
    }else if(algorithm == "AuxICA2"){
        num.iter <- if(is.null(num.iter)) 30 else num.iter
    }else if(algorithm == "IPCA"){
        num.iter <- if(is.null(num.iter)) 200 else num.iter
    }else if(algorithm == "SIMBEC"){
        num.iter <- if(is.null(num.iter)) 1000 else num.iter
        r_list <- if(is.null(r_list)) c(4) else r_list
        omega_for_each_r <- if(is.null(omega_for_each_r)) c(1.0) else omega_for_each_r
        a_r_for_each_r <- if(is.null(a_r_for_each_r)) c(1.0) else a_r_for_each_r
    }else if(algorithm == "AMUSE"){
        num.iter <- if(is.null(num.iter)) 1 else num.iter
        tau_list <- if(is.null(tau_list)) 1:50 else tau_list
    }else if(algorithm == "SOBI"){
        num.iter <- if(is.null(num.iter)) 100 else num.iter
        tau_list <- if(is.null(tau_list)) 1:10 else tau_list
    }else if(algorithm == "FOBI"){
        num.iter <- if(is.null(num.iter)) 100 else num.iter
    }else if(algorithm == "ProDenICA"){
        num.iter <- if(is.null(num.iter)) 100 else num.iter
        num_bins <- if(is.null(num_bins)) 20 else num_bins
    }else if(algorithm == "RICA"){
        num.iter <- if(is.null(num.iter)) 100 else num.iter
        alpha <- if(is.null(alpha)) 0.001 else alpha
        num_epoch <- if(is.null(num_epoch)) 100 else num_epoch
    }
    # Output
    list(X = X, J = J, algorithm = algorithm,
        num.iter = num.iter, thr = thr,
        r_list = r_list,
        omega_for_each_r = omega_for_each_r,
        a_r_for_each_r = a_r_for_each_r,
        tau_list = tau_list,
        num_bins = num_bins,
        alpha = alpha,
        num_epoch = num_epoch,
        verbose = verbose)
}

.initICA2_2 <- function(X, J, thr, allow_overcomplete){
    X_initialized <- .init_X(X, J)
    X <- X_initialized$X

    num_columns <- ncol(X)
    if(allow_overcomplete){
        W <- matrix(runif(num_columns * J), nrow = num_columns, ncol = J)
        A <- ginv(matrix(runif(J * num_columns), nrow = J, ncol = num_columns))
    }else{
        W <- matrix(runif(J * J), nrow = J, ncol = J)
        A <- ginv(W)
    }
    S <- X %*% W
    XBar <- S %*% A
    # Reconstruction Error
    RecError <- .recError(X, XBar)
    # Relative Change
    RelChange <- thr * 10
    # Output
    list(
        X = X, A = A, W = W, S = S,
        RecError = RecError, RelChange = RelChange,
        WhiteningMatrix = X_initialized$WhiteningMatrix,
        XMeans = X_initialized$XMeans, XSds = X_initialized$XSds
    )
}

.whitening2 <- function(X, scale = TRUE){
    num_sample <- nrow(X)
    X_scaled <- scale(X, scale = scale)
    eigen_X_scaled <- eigen(t(X_scaled) %*% X_scaled / num_sample)
    W <- eigen_X_scaled$vectors %*% diag(eigen_X_scaled$values^(-1 / 2)) %*% t(eigen_X_scaled$vectors)
    XWhiten <- X_scaled %*% W
    list(XWhiten = XWhiten, WhiteningMatrix = W)
}

.recError <- function(X, XBar){
    sum((X - XBar)^2) / sum(X^2)
}

.relChange <- function(iter, RecError){
    (RecError[iter - 1] - RecError[iter]) / RecError[iter - 1]
}

.FastICA2 <- function(X, W, S, J){
    n <- nrow(X)
    J <- ncol(X)
    # In order to conform to the notation of the paper,
    # W and X is treated as transposed.
    W <- t(W) # J * J
    X <- t(X) # J * n
    #########################################
    # Symmetric FastICA
    #########################################
    nonlinear_transform_function_name <- "tanh"
    # nonlinear_transform_function_name <- "exp"
    # nonlinear_transform_function_name <- "kurtosis"
    if(nonlinear_transform_function_name == "tanh"){
        alpha <- 1.0
        g <- function(x) tanh(alpha * x)
        gPrime <- function(x) alpha * (1 - tanh(alpha * x)^2)
    }else if(nonlinear_transform_function_name == "exp"){
        g <- function(x) x * exp(-x * 2 / 2)
        gPrime <- function(x) (1 - x^2) * exp(-x^2 / 2)
    }else if(nonlinear_transform_function_name == "kurtosis"){
        g <- function(x) x^3
        gPrime <- function(x) 3 * x^2
    }
    WX <- W %*% X # J * n
    gWX <- g(WX) # J * n
    gPrimeWX <- gPrime(WX) # J * n
    EgPrimeWX <- apply(gPrimeWX, 1, FUN = mean) # J
    EgPrimeWXW <- W %*% diag(EgPrimeWX) # J * J
    XgWX <- gWX %*% t(X) / n # J * J
    WNew <- XgWX - EgPrimeWXW # J * J
    # Orthgonalize
    WNewSvd <- svd(WNew)
    WNew <- tcrossprod(WNewSvd$u, WNewSvd$v)
    # Update
    t(WNew)
}

.ica_function_list <- list(
    "JADE" = .JADE,
    "AuxICA1" = .AuxICA1,
    "AuxICA2" = .AuxICA2,
    "IPCA" = .IPCA,
    "SIMBEC" = .SIMBEC,
    "AMUSE" = .AMUSE,
    "SOBI" = .SOBI,
    "FOBI" = .FOBI,
    "ProDenICA" = .ProDenICA,
    "RICA" = .RICA
)

# Simulation Dataset
X <- matrix(runif(100 * 50), nrow = 100, ncol = 50)

## Perform ICA against Simulation Dataset
J <- 5
out.JADE <- ICA2(X, J = J, algorithm = "JADE", verbose=TRUE)
out.AuxICA1 <- ICA2(X, J = J, algorithm = "AuxICA1", verbose=TRUE)
out.AuxICA2 <- ICA2(X, J = J, algorithm = "AuxICA2", verbose=TRUE)
out.IPCA <- ICA2(X, J = J, algorithm = "IPCA", verbose=TRUE)
out.SIMBEC <- ICA2(X, J = J, algorithm = "SIMBEC", verbose=TRUE)
out.AMUSE <- ICA2(X, J = J, algorithm = "AMUSE", verbose=TRUE)
out.SOBI <- ICA2(X, J = J, algorithm = "SOBI", verbose=TRUE)
out.FOBI <- ICA2(X, J = J, algorithm = "FOBI", verbose=TRUE)
out.ProDenICA <- ICA2(X, J = J, algorithm = "ProDenICA",
    num.iter = 1, verbose=TRUE)
out.RICA <- ICA2(X, J = J, algorithm = "RICA",
    num.iter = 1, verbose=TRUE)

## Test Output object / type
### Test O-1: Object
expect_identical(is.list(out.JADE), TRUE)
expect_identical(is.list(out.AuxICA1), TRUE)
expect_identical(is.list(out.AuxICA2), TRUE)
expect_identical(is.list(out.IPCA), TRUE)
expect_identical(is.list(out.SIMBEC), TRUE)
expect_identical(is.list(out.AMUSE), TRUE)
expect_identical(is.list(out.SOBI), TRUE)
expect_identical(is.list(out.FOBI), TRUE)
expect_identical(is.list(out.ProDenICA), TRUE)
expect_identical(is.list(out.RICA), TRUE)

### Test O-2: Object Names
expect_identical(
    sort(names(out.JADE)),
    sort(c("A", "W", "S", "X", "J", "algorithm", "num.iter", "thr", "verbose", "RecError", "RelChange")))
expect_identical(
    sort(names(out.AuxICA1)),
    sort(c("A", "W", "S", "X", "J", "algorithm", "num.iter", "thr", "verbose", "RecError", "RelChange")))
expect_identical(
    sort(names(out.AuxICA2)),
    sort(c("A", "W", "S", "X", "J", "algorithm", "num.iter", "thr", "verbose", "RecError", "RelChange")))
expect_identical(
    sort(names(out.IPCA)),
    sort(c("A", "W", "S", "X", "J", "algorithm", "num.iter", "thr", "verbose", "RecError", "RelChange")))
expect_identical(
    sort(names(out.SIMBEC)),
    sort(c("A", "W", "S", "X", "J", "algorithm", "num.iter", "thr", "verbose", "RecError", "RelChange")))
expect_identical(
    sort(names(out.AMUSE)),
    sort(c("A", "W", "S", "X", "J", "algorithm", "num.iter", "thr", "verbose", "RecError", "RelChange")))
expect_identical(
    sort(names(out.SOBI)),
    sort(c("A", "W", "S", "X", "J", "algorithm", "num.iter", "thr", "verbose", "RecError", "RelChange")))
expect_identical(
    sort(names(out.FOBI)),
    sort(c("A", "W", "S", "X", "J", "algorithm", "num.iter", "thr", "verbose", "RecError", "RelChange")))
expect_identical(
    sort(names(out.ProDenICA)),
    sort(c("A", "W", "S", "X", "J", "algorithm", "num.iter", "thr", "verbose", "RecError", "RelChange")))
expect_identical(
    sort(names(out.RICA)),
    sort(c("A", "W", "S", "X", "J", "algorithm", "num.iter", "thr", "verbose", "RecError", "RelChange")))

### Test 0-3: A
expect_identical(is.matrix(out.JADE$A), TRUE)
expect_identical(is.matrix(out.AuxICA1$A), TRUE)
expect_identical(is.matrix(out.AuxICA2$A), TRUE)
expect_identical(is.matrix(out.IPCA$A), TRUE)
expect_identical(is.matrix(out.SIMBEC$A), TRUE)
expect_identical(is.matrix(out.AMUSE$A), TRUE)
expect_identical(is.matrix(out.SOBI$A), TRUE)
expect_identical(is.matrix(out.FOBI$A), TRUE)
expect_identical(is.matrix(out.ProDenICA$A), TRUE)
expect_identical(is.matrix(out.RICA$A), TRUE)

# Except for IPCA, A is a square matrix because ICA is performed after decorrelating.
expect_equal(dim(out.JADE$A), c(J, J))
expect_equal(dim(out.AuxICA1$A), c(J, J))
expect_equal(dim(out.AuxICA2$A), c(J, J))
expect_equal(dim(out.IPCA$A), c(J, ncol(X)))
expect_equal(dim(out.SIMBEC$A), c(J, J))
expect_equal(dim(out.AMUSE$A), c(J, J))
expect_equal(dim(out.SOBI$A), c(J, J))
expect_equal(dim(out.FOBI$A), c(J, J))
expect_equal(dim(out.ProDenICA$A), c(J, J))
expect_equal(dim(out.RICA$A), c(J, J))

### Test 0-4: S
expect_identical(is.matrix(out.JADE$S), TRUE)
expect_identical(is.matrix(out.AuxICA1$S), TRUE)
expect_identical(is.matrix(out.AuxICA2$S), TRUE)
expect_identical(is.matrix(out.IPCA$S), TRUE)
expect_identical(is.matrix(out.SIMBEC$S), TRUE)
expect_identical(is.matrix(out.AMUSE$S), TRUE)
expect_identical(is.matrix(out.SOBI$S), TRUE)
expect_identical(is.matrix(out.FOBI$S), TRUE)
expect_identical(is.matrix(out.ProDenICA$S), TRUE)
expect_identical(is.matrix(out.RICA$S), TRUE)

expect_equal(dim(out.JADE$S), c(nrow(X), J))
expect_equal(dim(out.AuxICA1$S), c(nrow(X), J))
expect_equal(dim(out.AuxICA2$S), c(nrow(X), J))
expect_equal(dim(out.IPCA$S), c(nrow(X), J))
expect_equal(dim(out.SIMBEC$S), c(nrow(X), J))
expect_equal(dim(out.AMUSE$S), c(nrow(X), J))
expect_equal(dim(out.SOBI$S), c(nrow(X), J))
expect_equal(dim(out.FOBI$S), c(nrow(X), J))
expect_equal(dim(out.ProDenICA$S), c(nrow(X), J))
expect_equal(dim(out.RICA$S), c(nrow(X), J))

### Test 0-5: J
expect_identical(out.JADE$J, J)
expect_identical(out.AuxICA1$J, J)
expect_identical(out.AuxICA2$J, J)
expect_identical(out.IPCA$J, J)
expect_identical(out.SIMBEC$J, J)
expect_identical(out.AMUSE$J, J)
expect_identical(out.SOBI$J, J)
expect_identical(out.FOBI$J, J)
expect_identical(out.ProDenICA$J, J)
expect_identical(out.RICA$J, J)

### Test 0-6: algorithm
expect_identical(out.JADE$algorithm, "JADE")
expect_identical(out.AuxICA1$algorithm, "AuxICA1")
expect_identical(out.AuxICA2$algorithm, "AuxICA2")
expect_identical(out.IPCA$algorithm, "IPCA")
expect_identical(out.SIMBEC$algorithm, "SIMBEC")
expect_identical(out.AMUSE$algorithm, "AMUSE")
expect_identical(out.SOBI$algorithm, "SOBI")
expect_identical(out.FOBI$algorithm, "FOBI")
expect_identical(out.ProDenICA$algorithm, "ProDenICA")
expect_identical(out.RICA$algorithm, "RICA")

### Test 0-7: num.iter
# For iterative algorithm, num.iter matches set value of num.iter.
expect_identical(out.AuxICA1$num.iter, 30)
expect_identical(out.AuxICA2$num.iter, 30)

# For ProDenICA and RICA, num.iter are configured to 1.
expect_identical(out.ProDenICA$num.iter, 1)
expect_identical(out.RICA$num.iter, 1)

# For non-iterative algorithm, num.iter are configured to 1.
expect_identical(out.JADE$num.iter, 1)
expect_identical(out.IPCA$num.iter, 1)
expect_identical(out.SIMBEC$num.iter, 1)
expect_identical(out.AMUSE$num.iter, 1)
expect_identical(out.SOBI$num.iter, 1)
expect_identical(out.FOBI$num.iter, 1)

### Test 0-8: thr
expect_identical(out.JADE$thr, NULL)
expect_identical(out.AuxICA1$thr, formals(ICA)$thr)
expect_identical(out.AuxICA2$thr, formals(ICA)$thr)
expect_identical(out.IPCA$thr, NULL)
expect_identical(out.SIMBEC$thr, NULL)
expect_identical(out.AMUSE$thr, NULL)
expect_identical(out.SOBI$thr, NULL)
expect_identical(out.FOBI$thr, NULL)
expect_identical(out.ProDenICA$thr, formals(ICA)$thr)
expect_identical(out.RICA$thr, formals(ICA)$thr)

### Test 0-9: verbose
expect_identical(!out.JADE$verbose, formals(ICA)$verbose)
expect_identical(!out.AuxICA1$verbose, formals(ICA)$verbose)
expect_identical(!out.AuxICA2$verbose, formals(ICA)$verbose)
expect_identical(!out.IPCA$verbose, formals(ICA)$verbose)
expect_identical(!out.SIMBEC$verbose, formals(ICA)$verbose)
expect_identical(!out.AMUSE$verbose, formals(ICA)$verbose)
expect_identical(!out.SOBI$verbose, formals(ICA)$verbose)
expect_identical(!out.FOBI$verbose, formals(ICA)$verbose)
expect_identical(!out.ProDenICA$verbose, formals(ICA)$verbose)
expect_identical(!out.RICA$verbose, formals(ICA)$verbose)

### Test 0-10: RecError
expect_identical(is.vector(out.JADE$RecError), TRUE)
expect_identical(is.vector(out.AuxICA1$RecError), TRUE)
expect_identical(is.vector(out.AuxICA2$RecError), TRUE)
expect_identical(is.vector(out.IPCA$RecError), TRUE)
expect_identical(is.vector(out.SIMBEC$RecError), TRUE)
expect_identical(is.vector(out.AMUSE$RecError), TRUE)
expect_identical(is.vector(out.SOBI$RecError), TRUE)
expect_identical(is.vector(out.FOBI$RecError), TRUE)
expect_identical(is.vector(out.ProDenICA$RecError), TRUE)
expect_identical(is.vector(out.RICA$RecError), TRUE)

### Test O-11: RelChange
expect_identical(is.vector(out.AuxICA1$RelChange), TRUE)
expect_identical(is.vector(out.AuxICA2$RelChange), TRUE)
expect_identical(is.vector(out.ProDenICA$RelChange), TRUE)
expect_identical(is.vector(out.RICA$RelChange), TRUE)

## Test Error
### Test E-1: X
expect_error(ICA2(as.data.frame(X), J=J))

### Test E-2: J
expect_error(ICA2(X, J="5"))
expect_error(ICA2(X, J=c(2,4)))
expect_error(ICA2(X, J=10^10))

### Test E-3: num.iter
expect_error(ICA2(X, J=J, num.iter="100"))
expect_error(ICA2(X, J=J, num.iter=-1))

### Test E-4: thr
expect_error(ICA2(X, J=J, thr="0.1"))

### Test E-5: verbose
expect_error(ICA2(X, J=J, verbose="verbose"))

.ICA_Type1 <- function(){
    # Generate independent components.
    num_sample <- 500
    S_true <- cbind(
        runif(num_sample, min = -1, max = 1),
        runif(num_sample, min = -1, max = 1),
        runif(num_sample, min = -1, max = 1))
    # Get the number of elements.
    num_components <- ncol(S_true)
    # Generates a random mixing matrix.
    A_true <- matrix(
        runif(num_components^2, min = -1, max = 1),
        ncol = num_components)
    # Generate observed data.
    X_observed <- S_true %*% A_true
    list(X_observed=X_observed, S_true=S_true, A_true=A_true)
}

.ICA_Type2 <- function(){
    # Generate independent components.
    num_sample <- 500
    S_true <- cbind(
            rt(num_sample, df = 3),
            rt(num_sample, df = 3),
            rt(num_sample, df = 3)
        )
    num_components <- ncol(S_true)
    # Generate a random mixing matrix.
    A_true <- matrix(
        runif(num_components^2, min = -1, max = 1),
        ncol = num_components)
    X_observed <- S_true %*% A_true
    list(X_observed=X_observed, S_true=S_true, A_true=A_true)
}

.ICA_Type3 <- function(){
    # Generate independent components.
    num_sample <- 500
    S_true <- cbind(
            rt(num_sample, df = 1),
            rnorm(num_sample),
            runif(num_sample, min = -1, max = 1)
        )
    # Generate a random mixing matrix.
    num_components <- ncol(S_true)
    A_true <- matrix(
        runif(num_components^2, min = -1, max = 1),
        ncol = num_components)
    X_observed <- S_true %*% A_true
    list(X_observed=X_observed, S_true=S_true, A_true=A_true)
}

.ICA_Type4 <- function(){
    num_sample <- 500
    # Moving average window size.
    n_smooth <- 10
    # Generate independent components.
    S_true <- cbind(
      sin((1:num_sample) / 3),
      sin((1:num_sample) / 4),
      cos((1:num_sample) / 5))
    #  Smooth by moving average.
    for (i_var in 1:ncol(S_true)) {
      S_true[, i_var] <- filter(S_true[, i_var], rep(1, n_smooth)/n_smooth)
    }
    S_true <- S_true[(n_smooth + 1):(nrow(S_true) - n_smooth), ]
    num_sample <- nrow(S_true)
    num_components <- ncol(S_true)
    # Generate a random mixing matrix.
    A_true <- matrix(
        runif(num_components^2, min = -1, max = 1),
        ncol = num_components)
    X_observed <- S_true %*% A_true
    list(X_observed=X_observed, S_true=S_true, A_true=A_true)
}

.ICA_Type5 <- function(){
    env <- new.env()
    data(liver.toxicity, package="mixOmics", envir=env)
    env$liver.toxicity
}

.MICA <- function(){
    # Generates an observation vector of three mixed signals.
    # Read `5.Simulation` section in MICA paper.
    t_series <- seq(from = 0.00, to = 1.000, by = 1e-4)
    u1 <- runif(n = length(t_series), min = -1, max = 1)
    u2 <- sin(2 * pi * 800 * t_series + 6 * cos(2 * pi * 60 * t_series))
    u3 <- sin(2 * pi * 90 * t_series)
    u <- cbind(u3, u2, u1)
    v1 <- abs(u1)
    v2 <- abs(u2)
    v3 <- abs(u3)
    v <- cbind(v3, v2, v1)
    A_true <- matrix(runif(n = 3 * 3, min = -1, max = 1), nrow = 3, ncol = 3)
    B_true <- matrix(runif(n = 3 * 3, min = -1, max = 1), nrow = 3, ncol = 3)
    X <- u %*% A_true
    Y <- v %*% B_true
    list(X=X, Y=Y)
}

.GroupICA <- function(){
    # Generate data from a block-wise variance model
    d <- 6  # modified from 2 to 6 because Calhoun2009 with JADE does not work with d=2
    m <- 10
    n <- 5000
    group_index <- rep(c(1,2), each=n)
    partition_index <- rep(rep(1:m, each=n/m), 2)
    S <- matrix(NA, 2*n, d)
    H <- matrix(NA, 2*n, d)
    for(i in unique(group_index)){
      varH <- abs(rnorm(d))/4
      H[group_index==i, ] <- matrix(rnorm(d*n)*rep(varH, each=n), n, d)
      for(j in unique(partition_index[group_index==i])){
        varS <- abs(rnorm(d))
        index <- partition_index==j & group_index==i
        S[index,] <- matrix(rnorm(d*n/m)*rep(varS, each=n/m), n/m, d)
      }
    }
    A <- matrix(rnorm(d^2), d, d)
    A <- A%*%t(A)
    X <- t(A%*%t(S + H))
    Xs <- list()
    for (i_group in unique(group_index)) {
        Xs[[i_group]] <- X[group_index==i_group, ]
    }
    list(X=Xs[[1]], Y=Xs[[2]])
}

.flist <- list(
    ICA_Type1 = .ICA_Type1,
    ICA_Type2 = .ICA_Type2,
    ICA_Type3 = .ICA_Type3,
    ICA_Type4 = .ICA_Type4,
    ICA_Type5 = .ICA_Type5,
    MICA = .MICA,
    GroupICA = .GroupICA
)

#' Toy model data for using ICA, MICA, and GroupICA
#' There are 7 types of simulation:
#' ICA_Type1: Time-independent sub-gaussian data
#' ICA_Type2: Time-independent super-gaussian data
#' ICA_Type3: Data mixed with signals having no time dependence and different kurtosis
#' ICA_Type4: Time-dependent data
#' ICA_Type5: Toydata to model IPCA in N < P systems
#' MICA: Two time-serices data to model MICA
#' GroupICA: Toydata to model GroupICA
#' @param model "ICA_Type1", "ICA_Type2", "ICA_Type3", "ICA_Type4",
#' and "ICA_Type5", "MICA", and "GrouICA" are available
#' (Default: "ICA_Type1").
#' @param seeds Random number for setting set.seeds in the function (Default: 123).
#' @return A list containing simulation data sets.
#' @examples
#' data1 <- toyModel("ICA_Type1")
#' data2 <- toyModel("ICA_Type2")
#' data3 <- toyModel("ICA_Type3")
#' data4 <- toyModel("ICA_Type4")
#' data5 <- toyModel("ICA_Type5")
#' data6 <- toyModel("MICA")
#' data7 <- toyModel("GroupICA")
#' @importFrom utils data
#' @importFrom stats runif rt filter
#' @import mixOmics
#' @export
toyModel <- function(model = "ICA_Type1", seeds=123){
    set.seed(seeds)
    out <- .flist[[model]]()
    set.seed(NULL)
    out
}

# Simulation Datasets
data1 <- toyModel("ICA_Type1", seeds=1)
data2 <- toyModel("ICA_Type2", seeds=2)
data3 <- toyModel("ICA_Type3", seeds=3)
data4 <- toyModel("ICA_Type4", seeds=4)
data5 <- toyModel("ICA_Type5", seeds=5)
data6 <- toyModel("MICA", seeds=6)
data7 <- toyModel("GroupICA", seeds=7)

set.seed(1)
data1_ <- iTensor:::.ICA_Type1()
set.seed(2)
data2_ <- iTensor:::.ICA_Type2()
set.seed(3)
data3_ <- iTensor:::.ICA_Type3()
set.seed(4)
data4_ <- iTensor:::.ICA_Type4()
set.seed(5)
data5_ <- iTensor:::.ICA_Type5()
set.seed(6)
data6_ <- iTensor:::.MICA()
set.seed(7)
data7_ <- iTensor:::.GroupICA()

expect_true(length(data1) == 3)
expect_true(all(lapply(data1, nrow) == c(500,500,3)))
expect_true(all(lapply(data1, ncol) == 3))
expect_identical(data1, data1_)

expect_true(length(data2) == 3)
expect_true(all(lapply(data2, nrow) == c(500,500,3)))
expect_true(all(lapply(data2, ncol) == 3))
expect_identical(data2, data2_)

expect_true(length(data3) == 3)
expect_true(all(lapply(data3, nrow) == c(500,500,3)))
expect_true(all(lapply(data3, ncol) == 3))
expect_identical(data3, data3_)

expect_true(length(data4) == 3)
expect_true(all(lapply(data4, nrow) == c(480,480,3)))
expect_true(all(lapply(data4, ncol) == 3))
expect_identical(data4, data4_)

expect_true(length(data5) == 4)
expect_true(all(lapply(data5, nrow) == c(64,64,64,3116)))
expect_true(all(lapply(data5, ncol) == c(3116,10,4,2)))
expect_identical(data5, data5_)

expect_true(length(data6) == 2)
expect_true(all(lapply(data6, nrow) == 10001))
expect_true(all(lapply(data6, ncol) == 3))
expect_identical(data6, data6_)

expect_true(length(data7) == 2)
expect_true(all(lapply(data7, nrow) == 5000))
expect_true(all(lapply(data7, ncol) == 6))
expect_identical(data7, data7_)

##############################################
## Internal functions from groupICA package
## (Pfister, N. and Weichwald, S.)
## Original package: https://github.com/sweichwald/groupICA-R
## License: AGPL-3
## These functions are included here because the groupICA package
## is no longer maintained on CRAN.
##############################################

#' @importFrom MASS ginv
#' @importFrom stats cor

.groupICA_Pfister <- function(X,
                     group_index=NA,
                     partition_index=NA,
                     n_components=NA,
                     n_components_uwedge=NA,
                     rank_components=FALSE,
                     pairing='complement',
                     groupsize=1,
                     partitionsize=NA,
                     max_iter=1000,
                     tol=1e-12,
                     silent=TRUE){

  d <- dim(X)[2]
  n <- dim(X)[1]

  # generate group and partition indices as needed
  if(!is.numeric(group_index) & is.na(groupsize)){
    group_index <- rep(0, n)
  }
  else if(!is.numeric(group_index)){
    group_index <- .rigidgroup(n, groupsize)
  }
  if(!is.numeric(partition_index) & is.na(partitionsize)){
    smallest_group <- min(unique(group_index, return_counts=TRUE)$counts)
    partition_index <- .rigidpartition(group_index,
                                      max(c(d, floor(smallest_group/2))))
  }
  else if(!is.numeric(partition_index)){
    partition_index <- .rigidpartition(group_index, partitionsize)
  }


  no_groups <- length(unique(group_index))

  if(!silent){
    print('groupICA: computing covmats')
  }

  # estimate covariances (depending on pairing)
  if(pairing == 'complement'){
    no_pairs <- 0
    for(env in unique(group_index)){
      if(length(unique(partition_index[group_index == env]))>1){
        no_pairs <- no_pairs+length(unique(partition_index[group_index == env]))
      }
    }
    covmats <- vector("list", no_pairs)
    idx <- 1
    for(env in unique(group_index)){
      if(length(unique(partition_index[group_index == env]))==1){
        warning(paste("Removing group", toString(env),
                      "since the partition is trivial, i.e., contains only one set"))
      }
      else{
        for(subenv in unique(partition_index[group_index == env])){
          ind1 <- ((partition_index == subenv) &
                     (group_index == env))
          ind2 <- ((partition_index != subenv) &
                     (group_index == env))
          covmats[[idx]] <- cov(X[ind1,]) - cov(X[ind2,])
          idx <- idx + 1
        }
      }
    }
  }
  else if(pairing == 'allpairs'){
    no_pairs <-  0
    subvec <- rep(0, no_groups)
    for(i in 1:no_groups){
      env <- unique(group_index)[i]
      subvec[i] <- length(unique(partition_index[group_index == env]))
      no_pairs <- no_pairs + subvec[i]*(subvec[i]-1)/2
    }
    covmats <- vector("list", no_pairs)
    idx <- 1
    for(count in 1:no_groups){
      env <- unique(group_index)[count]
      unique_subs <- unique(partition_index[group_index == env])
      if(subvec[count] == 1){
        warning(paste("Removing group", toString(env),
                      "since the partition is trivial, i.e., contains only one set"))
      }
      else{
        for(i in 1:(subvec[count]-1)){
          for(j in (i+1):subvec[count]){
            ind1 <- ((partition_index == unique_subs[i]) &
                       (group_index == env))
            ind2 <- ((partition_index == unique_subs[j]) &
                       (group_index == env))
            covmats[[idx]] <- cov(X[ind1,]) - cov(X[ind2,])
            idx <- idx + 1
          }
        }
      }
    }
  }
  else{
    stop('no appropriate pairing specified')
  }

  # check if there are sufficiently many covariance matrices
  if(length(covmats)<=0){
    stop("Not sufficiently many covariance matrices.")
  }

  # add total observational covariance for normalization
  covmats <- append(list(cov(X)), covmats)

  if(!silent){
    print('groupICA: computed cov matrices')
  }

  # joint diagonalisation
  adj_res <- .uwedge_Pfister(covmats,
                    init=NA,
                    rm_x0=TRUE,
                    return_diag=FALSE,
                    tol=tol,
                    max_iter=max_iter,
                    n_components=n_components_uwedge,
                    silent=silent)
  V <- adj_res$V

  if(!silent){
    print('groupICA: finished uwedge ajd')
  }

  # rank components
  if(rank_components | !is.na(n_components)){
    A <- ginv(V)
    colcorrs <- rep(0, dim(V)[1])
    # running average
    for(k in 1:length(covmats)){
      colcorrs <- colcorrs + diag(abs(cor(A, covmats[[k]] %*% t(V)))) / length(covmats)
    }
    sorting <- order(colcorrs, decreasing=FALSE)
    V <- V[sorting,]
  }
  if(!is.na(n_components)){
    V <- V[1:n_components,]
  }

  # source estimation
  Shat <- t(V %*% t(X))


  return(list(V=V,
              Shat=Shat,
              converged=adj_res$converged,
              n_iter=adj_res$iteration,
              meanoffdiag=adj_res$meanoffdiag))
}


.rigidpartition <- function(group, nosamples){
  partition <- rep(0, length(group))
  for(e in unique(group)){
    partition[group == e] <- .rigidgroup(sum(group == e),
                                               nosamples) + max(partition)
  }
  return(partition)
}


.rigidgroup <- function(len, nosamples){
  groups <- floor(len / nosamples)
  changepoints <- round(seq(1, len, length.out=groups+1))
  index <- rep(0, len)
  for(i in 1:(length(changepoints)-1)){
    index[changepoints[i]:changepoints[i+1]] <- i
  }
  return(index)
}


.uwedge_Pfister <- function(Rx,
                   init=NA,
                   rm_x0=TRUE,
                   return_diag=FALSE,
                   tol=1e-10,
                   max_iter=1000,
                   n_components=NA,
                   silent=TRUE){


  # 0) Preprocessing

  # Remove and remember 0st matrix
  Rx0 <- Rx[[1]]
  if(rm_x0){
    Rx <- Rx[-1]
  }
  d <- dim(Rx0)[1]
  M <- length(Rx)

  if(is.na(n_components)){
    n_components <- d
  }

  # Initial guess
  if(!is.matrix(init) & n_components == d){
    EH <- eigen(Rx[[1]], symmetric=TRUE)
    V <- diag(1/sqrt(abs(EH$values))) %*% t(EH$vectors)
  }
  else if(!is.matrix(init)){
    EH <- eigen(Rx[[1]], symmetric=TRUE)
    mat <- matrix(0, n_components, d)
    diag(mat) <- 1/sqrt(abs(EH$values))[1:n_components]
    V <- mat %*% t(EH$vectors)
  }
  else{
    V <- init[1:n_components,]
  }

  V <- V / matrix(sqrt(rowSums(V^2)), n_components, d)

  converged <- FALSE
  for(iteration in 1:max_iter){

    # 1) Generate Rs
    Rs <- lapply(Rx, function(x) V %*% x %*% t(V))

    # 2) Use Rs to construct A, equation (24) in paper with W=Id
    # 3) Set A1=Id and substitute off-diagonals
    Rsdiag <- sapply(Rs, diag)
    Rsdiagprod <- Rsdiag %*% t(Rsdiag)
    denom_mat <- outer(diag(Rsdiagprod), diag(Rsdiagprod)) - Rsdiagprod^2
    Rkl_list = lapply(Rs, function(x) matrix(diag(x), n_components, n_components, byrow=T)*x)
    Rkl <- Reduce("+", Rkl_list)/M
    num_mat <- matrix(diag(Rsdiagprod), n_components, n_components)*Rkl - Rsdiagprod*t(Rkl)
    denom_mat[abs(denom_mat) < .Machine$double.tol] <- .Machine$double.tol
    diag(denom_mat) <- 1
    A <- num_mat / denom_mat
    diag(A) <- 1

    # 4) Set new V
    Vold <- V
    V <- qr.solve(A, Vold)

    # 5) Normalise V
    V <- V / matrix(sqrt(rowSums(V^2)), n_components, d)

    # 6) Check convergence
    changeinV = max(abs(V-Vold))
    if(changeinV < tol){
      converged = TRUE
      break
    }
  }

  # Rescale
  normaliser <- diag(V %*% Rx0 %*% t(V))
  V <- V / matrix((sign(normaliser)*sqrt(abs(normaliser))), n_components, d)
  Rxdiag <- lapply(Rx, function(x) V %*% x %*% t(V))
  entries_tot <- M*(n_components^2-n_components)
  meanoffdiag <- sqrt(sum(sapply(Rxdiag, function(x) sum(x^2)-sum(diag(x^2))))/entries_tot)

  # Return
  if(return_diag){
    res <- list(V=V,
                Rxdiag=Rxdiag,
                converged=converged,
                iteration=iteration,
                meanoffdiag=meanoffdiag)
  }
  else{
    res <- list(V=V,
                Rxdiag=NA,
                converged=converged,
                iteration=iteration,
                meanoffdiag=meanoffdiag)
  }

  return(res)
}

---
title: "2. Multimodal Independent Component Analysis (MICA) and Group Independent Component Analysis (GroupICA)"
author:
- name: Koki Tsuyuzaki
  affiliation: Department of Artificial Intelligence Medicine,
    Graduate School of Medicine, Chiba University
  email: k.t.the-answer@hotmail.co.jp
date: "`r Sys.Date()`"
bibliography: bibliography.bib
package: iTensor
output: rmarkdown::html_vignette
vignette: |
  %\VignetteIndexEntry{2. Multimodal Independent Component Analysis (MICA) and Group Independent Component Analysis (GroupICA)}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

# Introduction

In this vignette we consider approximating multiple data matrices as a product of multiple low-rank matrices (a.k.a., factor matrices).

Test data is available from `toyModel`.

```{r data, echo=TRUE}
library("iTensor")
data1 <- iTensor::toyModel("MICA")
data2 <- iTensor::toyModel("GroupICA")
str(data1, 2)
str(data2, 2)
```

Both `data1` and `data2` contain two time-series data `X` and `Y` as follows.

```{r data2, echo=TRUE, fig.height=4, fig.width=4}
plot.ts(data1$X[7700:8000,], main="data1 (X)")
plot.ts(data1$Y[7700:8000,], main="data1 (Y)")
```

```{r data3, echo=TRUE, fig.height=4, fig.width=8}
plot.ts(data2$X[4700:5000,], main="data2 (X)")
plot.ts(data2$Y[4700:5000,], main="data2 (Y)")
```

# Multimodal Independent Component Analysis (MICA)

As a formulation that extends ICA (independent component analysis) to the multiple matrices case, Multimodal Independent Component Analysis (MICA) was proposed ([@mica]). MICA extracts statistically dependent pairs of features from the sources, where the components of feature vector extracted from each source are independent.

`MICA` can be performed as follows.

```{r mica, echo=TRUE, fig.height=4, fig.width=4}
t_series <- seq(from = 0.00, to = 1.000, by = 1e-4)
out.MICA <- MICA(data1$X, data1$Y, J=3, gamma_ts = 1 - 1 / (1 + exp(-100 * (t_series - 0.3))))
```

`J` is the rank parameter for ICA and `gamma_ts` is the weighting factor for dependence on independence.
You will see that `MICA` could extract some time-series signals.

```{r pairs_mica, echo=TRUE, fig.height=4, fig.width=4}
plot.ts(out.MICA$U[7700:8000, ], main="Source Signal (X)")
plot.ts(out.MICA$V[7700:8000, ], main="Source Signal (Y)")
```

# Group Independent Component Analysis (GroupICA)

Another formulation of the decomposition is Group Independent Component Analysis (GroupICA [@groupica1; @groupica2]). `GroupICA` can be performed as follows.

```{r groupica, echo=TRUE}
out_groupica <- GroupICA(data2, J1=6, algorithm="pooled")
```

The rank for each factor matrix can be set as `J1` and the decomposition algorithm can be easily switched by `ica.algorithm` (`algorithm` of `ICA` and `ICA2` can be specified). To pool the results of ICA against each data matrix, we implemented three algorithms such as `pooled`, `Calhoun2009`, and `Pfister2018`. For the details, see the references [@groupica1; @groupica2].

You will see that `GroupICA` could extract some time-series signals.

```{r plot_groupica, echo=TRUE, fig.height=4, fig.width=8}
plot.ts(out_groupica$Ss[[1]], main="Source Signal (X)")
plot.ts(out_groupica$Ss[[2]], main="Source Signal (Y)")
```

Unlike `MICA`, `GroupICA` can also be applied to more than three matrices.

# Session Information {.unnumbered}

```{r sessionInfo, echo=FALSE}
sessionInfo()
```

# References

---
title: "1. Independent Component Analysis (ICA)"
author:
- name: Koki Tsuyuzaki
  affiliation: Department of Artificial Intelligence Medicine,
    Graduate School of Medicine, Chiba University
  email: k.t.the-answer@hotmail.co.jp
date: "`r Sys.Date()`"
bibliography: bibliography.bib
package: iTensor
output: rmarkdown::html_vignette
vignette: |
  %\VignetteIndexEntry{1. Independent Component Analysis (ICA)}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

# Introduction

In this vignette, we consider approximating a data matrix as a product of multiple low-rank matrices (a.k.a., factor matrices).

Test data is available from `toyModel`.

```{r data, echo=TRUE}
library("iTensor")
data1 <- iTensor::toyModel("ICA_Type1")
data2 <- iTensor::toyModel("ICA_Type2")
data3 <- iTensor::toyModel("ICA_Type3")
data4 <- iTensor::toyModel("ICA_Type4")
data5 <- iTensor::toyModel("ICA_Type5")
dim(data1$X_observed)
dim(data2$X_observed)
dim(data3$X_observed)
dim(data4$X_observed)
dim(data5$gene) # N < P systems
```

Summary of these data is as follows:

1. ICA with time-independent sub-Gaussian data
2. ICA with time-independent super-Gaussian data
3. ICA with data mixed with signals having no time dependence and different kurtosis
4. ICA with time-dependent data
5. IPCA in N < P systems

You will see that the stuctures of these data as follows.

```{r data2, echo=TRUE, fig.height=4, fig.width=4}
pairs(data1$X_observed, main="data1")
pairs(data2$X_observed, main="data2")
pairs(data3$X_observed, main="data3")
pairs(data4$X_observed, main="data4")
dim(data5$gene)
```

To perform and visualize all the ICA at once, here we write some functions as follows.

```{r setting, echo=TRUE}
allICA <- function(X, J){
    # Classic ICA Methods
    out.FastICA <- ICA(X=X, J=J, algorithm="FastICA")
    out.InfoMax <- ICA(X=X, J=J, algorithm="InfoMax")
    out.ExtInfoMax <- ICA(X=X, J=J, algorithm="ExtInfoMax")
    # Modern ICA Method
    out.JADE <- ICA2(X=X, J=J, algorithm="JADE")
    # Output
    list(out.FastICA=out.FastICA, out.InfoMax=out.InfoMax,
      out.ExtInfoMax=out.ExtInfoMax, out.JADE=out.JADE)
}

CorrIndexAllICA <- function(S, out){
    tmp <- lapply(out, function(x, S){CorrIndex(cor(S, Re(x$S)))}, S=S)
    Name <- gsub("out.", "", names(tmp))
    Value <- round(unlist(tmp), 3)
    names(Value) <- NULL
    cbind(Name, Value)
    tmp <- as.data.frame(cbind(Name, Value))
    colnames(tmp) <- c("Method", "CorrIndex")
    knitr::kable(tmp)
}
```

Note that we select only four ICA algorithms for the demonstration but in `iTensor` 12 ICA algorithms are available. For the details, check `?ICA` and `?ICA2`.

# 1. ICA with time-independent sub-Gaussian data

In this data, according to the independence between estimated source signal, an upright square means that the calculation is successful and other cases such as a rhombus rotated 45 degrees from a square means that the calculation is fail.

```{r ica1, echo=TRUE}
set.seed(123456)
out1 <- allICA(X=data1$X_observed, J=3)
```

Here we use CorrIndex, which is a performance index to evaluate ICA results [@corrindex]. CorrIndex imply that FastICA, ExtInfoMax, and JADE performed well (the closer to 0, the better the performance).

```{r corrindex1, echo=TRUE}
CorrIndexAllICA(data1$S_true, out1)
```

We can see the details of source signals as follows.

```{r pairs_ica1, echo=TRUE, fig.height=4, fig.width=4}
pairs(out1$out.FastICA$S, main="FastICA")
pairs(out1$out.InfoMax$S, main="InfoMax")
pairs(out1$out.ExtInfoMax$S, main="ExtInfoMax")
pairs(Re(out1$out.JADE$S), main="JADE")
```

# 2. ICA with time-independent super-Gaussian data

In this data, if the outliers are distributed along the lines of x = 0 and y = 0, the calculation is successful and if they are of the cross-shape, it is a failure.

```{r ica2, echo=TRUE}
set.seed(123456)
out2 <- allICA(X=data2$X_observed, J=3)
```

CorrIndex imply that all the algorithms performed well.

```{r corrindex2, echo=TRUE}
CorrIndexAllICA(data2$S_true, out2)
```

We can see the details of source signals as follows.

```{r pairs_ica2, echo=TRUE, fig.height=4, fig.width=4}
pairs(out2$out.FastICA$S, main="FastICA")
pairs(out2$out.InfoMax$S, main="InfoMax")
pairs(out2$out.ExtInfoMax$S, main="ExtInfoMax")
pairs(Re(out2$out.JADE$S), main="JADE")
```

# 3. ICA with data mixed with signals having no time dependence and different kurtosis

In this data, the uniform distribution should be extracted in such a way that it is not rhombic, and furthermore the normal distribution and the t-distribution with 1 degree of freedom should be extracted independently. Sometimes, it is difficult ot determine visually though.

```{r ica3, echo=TRUE}
set.seed(123456)
out3 <- allICA(X=data3$X_observed, J=3)
```

CorrIndex imply that FastICA, ExtInfoMax, and JADE performed well.

```{r corrindex3, echo=TRUE}
CorrIndexAllICA(data3$S_true, out3)
```

We can see the details of source signals as follows.

```{r pairs_ica3, echo=TRUE, fig.height=4, fig.width=4}
pairs(out3$out.FastICA$S, main="FastICA")
pairs(out3$out.InfoMax$S, main="InfoMax")
pairs(out3$out.ExtInfoMax$S, main="ExtInfoMax")
pairs(Re(out3$out.JADE$S), main="JADE")
```

# 4. ICA with time-dependent data

In this data, if the mesh pattern is reproduced, the calculation is successful.

```{r ica4, echo=TRUE}
set.seed(123456)
out4 <- allICA(X=data4$X_observed, J=3)
```

CorrIndex imply that FastICA, ExtInfoMax, and JADE performed well.

```{r corrindex4, echo=TRUE}
CorrIndexAllICA(data4$S_true, out4)
```

We can see the details of source signals as follows.

```{r pairs_ica4, echo=TRUE, fig.height=4, fig.width=4}
pairs(out4$out.FastICA$S, main="FastICA")
pairs(out4$out.InfoMax$S, main="InfoMax")
pairs(out4$out.ExtInfoMax$S, main="ExtInfoMax")
pairs(Re(out4$out.JADE$S), main="JADE")
```

# 5. IPCA in N < P systems

This kind of data is an thin and long matrix, which is a very common data structure in omics data. In `iTensor`, we re-implemented the IPCA function of `mixOmics` package.

```{r ica5, echo=TRUE}
library("mixOmics")
set.seed(123456)

# mixOmics's IPCA
res.ipca <- ipca(data5$gene, ncomp=3, mode="deflation")

# iTensor's IPCA
out5 <- ICA2(X=as.matrix(data5$gene), J=3, algorithm="IPCA")
```

We can see the details of source signals as follows. In this data, we can confirm that the IPCA of `mixOmics` and `iTensor` have the same results, except for the order and constant-fold relationships.

```{r pairs_ica5, echo=TRUE, fig.height=4, fig.width=4}
pairs(res.ipca$x, main="IPCA (mixOmics)", col=data5$treatment[,"Treatment.Group"], pch=16)
pairs(out5$S, main="IPCA (iTensor)", col=data5$treatment[,"Treatment.Group"], pch=16)
```

# Session Information {.unnumbered}

```{r sessionInfo, echo=FALSE}
sessionInfo()
```

# References

---
title: "3. Multilinear Independent Component Analysis (MultilinearICA)"
author:
- name: Koki Tsuyuzaki
  affiliation: Department of Artificial Intelligence Medicine,
    Graduate School of Medicine, Chiba University
  email: k.t.the-answer@hotmail.co.jp
date: "`r Sys.Date()`"
bibliography: bibliography.bib
package: iTensor
output: rmarkdown::html_vignette
vignette: |
  %\VignetteIndexEntry{3. Multilinear Independent Component Analysis (MultilinearICA)}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

# Introduction

In this vignette we consider approximating a tensor as a product of multiple low-rank matrices (a.k.a., factor matrices) and a core tensor.

Test data is available from `toyModel`.

```{r data, echo=TRUE}
library("iTensor")
library("nnTensor")
data <- nnTensor::toyModel("CP")
str(data, 2)
```

You will see that there are four small blocks in the diagonal direction of the data tensor.

```{r data2, echo=TRUE, fig.height=6, fig.width=6}
plotTensor3D(data)
```

# Multilinear Independent Component Analysis (MultilinearICA)

There are so many tensor decomposition algorithms but here we consider that each factor matrix is extracted by Independent Component Analysis (ICA). This is called Multilinear Independent Component Analysis (MultilinearICA [@multilinearica]).

`MultilinearICA` can be performed as follows.

```{r multilinearica, echo=TRUE}
out <- MultilinearICA(data, Js=c(4,4,4), algorithm="FastICA")
```

The rank for each factor matrix can be set as `Js` and the decomposition algorithm can be easily switched by `algorithm`.

By using `recTensor` of `nnTensor`, user can easily reconstruct the data from core tensor and factor matrices as follows.

```{r plot_multilinearica, echo=TRUE, fig.height=6, fig.width=6}
rec_data <- recTensor(out$S, out$As)
plotTensor3D(rec_data)
```

# Session Information {.unnumbered}

```{r sessionInfo, echo=FALSE}
sessionInfo()
```

# References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MultilinearICA.R
\name{MultilinearICA}
\alias{MultilinearICA}
\title{Multilinear independent component analysis}
\usage{
MultilinearICA(
  X,
  Js = c(3, 3, 3),
  modes = 1:3,
  algorithm = c("FastICA", "InfoMax", "ExtInfoMax")
)
}
\arguments{
\item{X}{An rTensor object}

\item{Js}{A vector to specify the rank in each mode (Default: c(3,3,3))}

\item{modes}{A vector to specify which modes are decomposed (Default: 1:3)}

\item{algorithm}{The algorithm to decompose the input tensor in each mode (Default: "FastICA")}
}
\value{
A list containing the result of the decomposition
}
\description{
#' The input object is assumed to be a Tensor object defined by rTensor package.
In MultilinearICA, ICA function is performed in each mode of the tensor.
}
\examples{
library("rTensor")
arrX <- array(runif(10*20*30), dim=c(10,20,30))
X <- as.tensor(arrX)
Js <- c(2,3,4)
out <- MultilinearICA(X, Js=Js)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GroupICA.R
\name{GroupICA}
\alias{GroupICA}
\title{Group Independent Component Analysis (GroupICA)}
\usage{
GroupICA(
  Xs,
  J1,
  J2 = J1,
  algorithm = c("pooled", "Calhoun2009", "Pfister2018"),
  ica.algorithm = c("FastICA", "InfoMax", "ExtInfoMax", "JADE", "AuxICA1", "AuxICA2",
    "IPCA", "SIMBEC", "AMUSE", "SOBI", "FOBI", "ProDenICA", "RICA"),
  num.iter = 30,
  thr = 1e-10,
  verbose = FALSE
)
}
\arguments{
\item{Xs}{A list containing multiple matrices}

\item{J1}{Rank parameter to decompose}

\item{J2}{Rank parameter used in Calhoun2009}

\item{algorithm}{Pool algorithm to merge multiple ICA results (Default: pooled)}

\item{ica.algorithm}{The decomposition algorithm (Default: "FastICA")}

\item{num.iter}{The number of iterations}

\item{thr}{The threshold to terminate the iteration (Default: 1E-10)}

\item{verbose}{Verbose option}
}
\value{
A list containing the result of the decomposition
}
\description{
The input data is assumed to be a list containing multiple matrices, which share common column.
}
\examples{
X1 <- matrix(runif(100*200), nrow=100, ncol=200)
X2 <- matrix(runif(150*200), nrow=150, ncol=200)
Xs <- list(X1=X1, X2=X2)
out <- GroupICA(Xs, J1=5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MICA.R
\name{MICA}
\alias{MICA}
\title{Multimodal independent component analysis}
\usage{
MICA(
  X,
  Y,
  J,
  eta = 1000 * 1e-04,
  verbose = FALSE,
  mu = 50 * 1e-04,
  gamma_ts = 1
)
}
\arguments{
\item{X}{A matrix sharing the column space with Y (??? x N)}

\item{Y}{A matrix sharing the column space with X (??? x N)}

\item{J}{The rank parameter to decompose the matrices}

\item{eta}{A learning rate parameter of stochastic gradient descent}

\item{verbose}{Verbose option}

\item{mu}{A learning rate parameter of stochastic gradient descent}

\item{gamma_ts}{Weighting factor for dependence on independence}
}
\value{
A list containing the result of the decomposition
}
\description{
The input datasets are assumed to be two matrices sharing the column space.
MICA decomposes the matrices simutanously
and extracts the components that maximizes the mutual information
between the components.
}
\examples{
X <- array(runif(10*20), dim=c(10,20))
Y <- array(runif(15*20), dim=c(15,20))
J <- 20
out <- MICA(X, Y, J=J)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CorrIndex.R
\name{CorrIndex}
\alias{CorrIndex}
\title{CorrIndex}
\usage{
CorrIndex(cross_correlation_matrix)
}
\arguments{
\item{cross_correlation_matrix}{Cross-correlation matrix}
}
\value{
CorrIndex, which means the closeness between S and S_true , is returned.
}
\description{
Calculate the CorrIndex of the cross-correlation matrix of S_true and estimated S.
The closer the value is to 0, the closer estimated S is to S_true.
}
\examples{
S_true <- matrix(runif(5*5), nrow=5, ncol=5)
S <- matrix(runif(5*5), nrow=5, ncol=5)
CorrIndex(cor(S_true, S))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/toyModel.R
\name{toyModel}
\alias{toyModel}
\title{Toy model data for using ICA, MICA, and GroupICA
There are 7 types of simulation:
ICA_Type1: Time-independent sub-gaussian data
ICA_Type2: Time-independent super-gaussian data
ICA_Type3: Data mixed with signals having no time dependence and different kurtosis
ICA_Type4: Time-dependent data
ICA_Type5: Toydata to model IPCA in N < P systems
MICA: Two time-serices data to model MICA
GroupICA: Toydata to model GroupICA}
\usage{
toyModel(model = "ICA_Type1", seeds = 123)
}
\arguments{
\item{model}{"ICA_Type1", "ICA_Type2", "ICA_Type3", "ICA_Type4",
and "ICA_Type5", "MICA", and "GrouICA" are available
(Default: "ICA_Type1").}

\item{seeds}{Random number for setting set.seeds in the function (Default: 123).}
}
\value{
A list containing simulation data sets.
}
\description{
Toy model data for using ICA, MICA, and GroupICA
There are 7 types of simulation:
ICA_Type1: Time-independent sub-gaussian data
ICA_Type2: Time-independent super-gaussian data
ICA_Type3: Data mixed with signals having no time dependence and different kurtosis
ICA_Type4: Time-dependent data
ICA_Type5: Toydata to model IPCA in N < P systems
MICA: Two time-serices data to model MICA
GroupICA: Toydata to model GroupICA
}
\examples{
data1 <- toyModel("ICA_Type1")
data2 <- toyModel("ICA_Type2")
data3 <- toyModel("ICA_Type3")
data4 <- toyModel("ICA_Type4")
data5 <- toyModel("ICA_Type5")
data6 <- toyModel("MICA")
data7 <- toyModel("GroupICA")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ICA2.R
\name{ICA2}
\alias{ICA2}
\title{Independent Component Analysis (Modern Methods)}
\usage{
ICA2(
  X,
  J,
  algorithm = c("JADE", "AuxICA1", "AuxICA2", "IPCA", "SIMBEC", "AMUSE", "SOBI",
    "FOBI", "ProDenICA", "RICA"),
  num.iter = NULL,
  thr = 1e-10,
  r_list = NULL,
  omega_for_each_r = NULL,
  a_r_for_each_r = NULL,
  tau_list = NULL,
  num_bins = NULL,
  alpha = NULL,
  num_epoch = NULL,
  verbose = FALSE
)
}
\arguments{
\item{X}{A matrix}

\item{J}{Rank parameter to decompose}

\item{algorithm}{The decomposition algorithm (Default: "JADE")}

\item{num.iter}{The number of iteration}

\item{thr}{The threshold to terminate the iteration (Default: 1E-10)}

\item{r_list}{List of r-th order cumulants used in SIMBEC (Default: NULL)}

\item{omega_for_each_r}{Weight vector of r_list used in SIMBEC (Default: NULL)}

\item{a_r_for_each_r}{Parameter vector to specify the shape of partial activation function in SIMBEC (Default: NULL)}

\item{tau_list}{List of lags to consider the auto-correlation used in AMUSE and SOBI (Default: NULL)}

\item{num_bins}{Number of bins for histgram in ProDenICA (Default: NULL)}

\item{alpha}{Learning rate used for gradient descent in RICA (Default: NULL)}

\item{num_epoch}{Number of epoch used for gradient descent in RICA (Default: NULL)}

\item{verbose}{Verbose option}
}
\value{
A list containing the result of the decomposition
}
\description{
The input data is assumed to be a matrix.
ICA decomposes the matrix and extract the components that
are statistically independent each other.
}
\examples{
ICA2
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ICA.R
\name{ICA}
\alias{ICA}
\title{Independent Component Analysis (Classic Methods)}
\usage{
ICA(
  X,
  J,
  algorithm = c("FastICA", "InfoMax", "ExtInfoMax"),
  num.iter = 100,
  thr = 1e-10,
  nonlinear_func = c("tanh", "exp", "kurtosis"),
  learning_rate = 1,
  verbose = FALSE
)
}
\arguments{
\item{X}{A matrix}

\item{J}{Rank parameter to decompose}

\item{algorithm}{The decomposition algorithm (Default: "FastICA")}

\item{num.iter}{The number of iteration}

\item{thr}{The threshold to terminate the iteration (Default: 1E-10)}

\item{nonlinear_func}{The function used in FastICA (Default: "tanh")}

\item{learning_rate}{The learning rate used in InfoMax or ExtInfoMax}

\item{verbose}{Verbose option}
}
\value{
A list containing the result of the decomposition
}
\description{
The input data is assumed to be a matrix.
ICA decomposes the matrix and extract the components that
are statistically independent each other.
}
\examples{
X <- matrix(runif(100*200), nrow=100, ncol=200)
J <- 5
out.FastICA <- ICA(X, J=J, algorithm="FastICA")
out.InfoMax <- ICA(X, J=J, algorithm="InfoMax")
out.ExtInfoMax <- ICA(X, J=J, algorithm="ExtInfoMax")
}
